#ifndef GOAPI_EVENT_H
#define GOAPI_EVENT_H

#include <GoApi/GoApiDef.h>
#include <functional>
#include <list>

namespace Go
{
    template <typename... Args>
    class Event
    {
    public:
        typedef std::function<void(Args...)> CallbackType;

        class CallbackId
        {
        public:
            CallbackId() : valid(false) {};

        private:
            friend class Event<Args...>;
            CallbackId(typename std::list<CallbackType>::iterator i)
                : iter(i), valid(true)
            {}

            typename std::list<CallbackType>::iterator iter;
            bool valid;
        };

        CallbackId AddListener(CallbackType callback)
        {
            if (callback)
            {
                callbackList.push_back(callback);
                return CallbackId(--callbackList.end());
            }

            return CallbackId();
        }

        void RemoveListener(CallbackId& id)
        {
            if (id.valid)
            {
                callbackList.erase(id.iter);
            }
        }

        void Notify(Args... args)
        {
            for (auto& callback : callbackList)
            {
                callback(args...);
            }
        }

        size_t Count()
        {
            return callbackList.size();
        }

        void ClearListeners()
        {
            callbackList.clear();
        }

    private:
        std::list<CallbackType> callbackList;
    };

    /*! @class BeginEndUpdater Event.h "GoApi/GoApi/Event.h"
     *  @brief Helper class for keeping tracking of events between a BeginUpdate()
     *         and EndUpdate() cycle.
     *
     *  @remarks  Example usage see below:
     *  @code {.cpp}
     *  class FooClass
     *  {
     *  public:
     *      FooClass()
     *      { 
     *      }
     *
     *      void UpdateSomePart()
     *      {
     *          OnUpdated();
     *      }
     *
     *      void UpdateSomeOtherPart()
     *      {
     *          OnUpdated();
     *      }
     *
     *      Go::Event<> UpdatedEvent()
     *      {
     *           return updatedEvent;
     *      }
     *
     *      void OnUpdated()
     *      {
     *          updater.SoftUpdate([this]() {
     *             updatedEvent.Notify();
     *          });
     *      }
     *
     *       void BeginUpdate()
     *      {
     *          updater.BeginUpdate();
     *      }
     *
     *      void EndUpdate()
     *      {
     *          updater.EndUpdate([this]() {
     *             updatedEvent.Notify();
     *          });
     *
     *      }
     * 
     *  private:
     *      Go::Event<> updatedEvent;
     *      BeginEndUpdater updater;
     *  };
     *  @endcode
     */
    class BeginEndUpdater
    {
    public:
        /**
         * Constructor.
         */
        BeginEndUpdater()
        {
            updating = false;
            updatedCount = 0;
        }

        ~BeginEndUpdater()
        {
        }

        /**
         * Begins an update cycle.
         */
        void BeginUpdate()
        {
            updating = true;
        }

        /**
         * Returns whether we are currengly within an update cycle.
         *
         * @return true if within an update cycle, false otherwise.
         */
        bool IsUpdating()
        {
            return updating;
        }

        /**
         * Performs an update if not during an update cycle, otherwise increments an internal counter
         * checked when @ref EndUpdate() is called.
         *
         * @param updateFunc Callback funciton to perform if not during an update cycle.
         */
        void SoftUpdate(std::function<void()> updateFunc)
        {
            if (updating)
            {
                // Within updating cycle, so remember the update.
                updatedCount++;
            }
            else
            {
                updateFunc();
            }
        }

        /**
         * Ends an update cycle without performing any update action.
         */
        void EndUpdate()
        {
            EndUpdate([](){});
        }

        /**
         * Performs an update when ending an update cycle, if internal counter was incremented
         * with @ref SoftUpdate().
         *
         * @param updateFunc Callback funciton to perform upon ending an update cycle.
         */
        void EndUpdate(std::function<void()> updateFunc)
        {
            // Update is ending.
            if (updating)
            {
                updating = false;
                if (updatedCount > 0)
                {
                    updatedCount = 0;
                    updateFunc();
                }
            }
        }

    private:
        bool updating;
        size_t updatedCount;
    };
}

#endif
