#ifndef GOAPI_DYNAMIC_PARAMS_H
#define GOAPI_DYNAMIC_PARAMS_H

#include <GoApi/Params/Params.h>

namespace Go {
namespace Params {

// GOS-6439 This is partial implmenetation of DynanmicValue.
// There needs to be additional work for related DynamicEmum, DynamicArray, etc.

// GOS-6452 This further improvement may deprecate the existing work
// done in this class.
template <typename T>
class DynamicValue : public Go::Properties::Value<T>, public ParamsBase
{
public:
    // GOS-5307: Save validated value to ensure read call 
    mutable T validatedValue;

    DynamicValue()
    {
        RegisterCommonAttrs();

        T minVal = this->Get();
        bool setMin = UnsignedValueMin<T>::SetUnsignedMin(minVal);
        if (setMin)
        {
            this->EnableMinimum().GetMinimum().Set(minVal);
        }
    }

    DynamicValue(const T& val) :
        Go::Properties::Value<T>(val)
    {
        RegisterCommonAttrs();

        T minVal = this->Get();
        bool setMin = UnsignedValueMin<T>::SetUnsignedMin(minVal);
        if (setMin)
        {
            this->EnableMinimum().GetMinimum().Set(minVal);
        }
    }

    ~DynamicValue()
    {
        if (units)
        {
            delete units;
        }

        if (description)
        {
            delete description;
        }

        if (actual)
        {
            delete actual;
        }

        if (minimum)
        {
            delete minimum;
        }

        if (maximum)
        {
            delete maximum;
        }

        if (exclusiveMinimum)
        {
            delete exclusiveMinimum;
        }

        if (exclusiveMaximum)
        {
            delete exclusiveMaximum;
        }

        if (scale)
        {
            delete scale;
        }

        if (extendedType)
        {
            delete extendedType;
        }
    }

    // Implementation of virtual method from IValueIo
    void ReadValue(Go::Properties::IValueReader& reader) override
    {
        // GOS-4877: Changed readOnly value doesn't fail validation anymore,
        // but still don't want to change the value.
        // GOS-5307: readOnly shouldn't prevent changes unless validation was previously run.
        if (readOnlyValidateFailed)
        {
            readOnlyValidateFailed = false;
            // If new value matches the one from validation, assume this is the same session.
            // This isn't air-tight, but helps prevent theoretical case where validation sets flag then later fails, 
            // then later a new value is read without validation (such as job load).
            T newValue;
            Go::Properties::ValueIoDispatch<T>::Read(reader, newValue);
            if (Go::Properties::ValueEqFunc<T>::Eq(newValue, validatedValue))
            {
                // Reject change -- readOnly failed validation.
                return;
            }
        }

        Go::Properties::Value<T>::ReadValue(reader);
    }

    DynamicValue<T>& operator=(const T& other)
    {
        this->Set(other);
        return *this;
    }

    void CopyValues(const Go::Properties::Node& other)
    {
        Go::Properties::Value<T>::CopyValues(other);

        auto& otherValue = dynamic_cast<const DynamicValue<T>&>(other);
        ParamsBase::CopyValues(otherValue);

        if (readOnlyEnabled)
        {
            EnableReadOnly();
        }

        if (usedEnabled)
        {
            EnableUsed();
        }

        if (otherValue.description)
        {
            EnableUnits();
            units->Set(otherValue.units->Get());
        }

        if (otherValue.description)
        {
            EnableDescription();
            description->Set(otherValue.description->Get());
        }

        if (otherValue.actual)
        {
            EnableActual();
            actual->Set(otherValue.actual->Get());
        }

        if (otherValue.minimum)
        {
            EnableMinimum();
            minimum->Set(otherValue.minimum->Get());
        }

        if (otherValue.maximum)
        {
            EnableMaximum();
            maximum->Set(otherValue.maximum->Get());
        }

        if (otherValue.exclusiveMinimum)
        {
            EnableExclusiveMinimum();
            exclusiveMinimum->Set(otherValue.exclusiveMinimum->Get());
        }

        if (otherValue.exclusiveMaximum)
        {
            EnableExclusiveMaximum();
            exclusiveMaximum->Set(otherValue.exclusiveMaximum->Get());
        }

        if (otherValue.scale)
        {
            EnableScale();
            scale->Set(otherValue.scale->Get());
        }

        if (otherValue.extendedType)
        {
            SetExtendedType(otherValue.extendedType->Get());
        }
    }

    void CheckValidation(const T& newValue) const override
    {
        ParamsBase::CheckValidation();
        if (readOnlyValidateFailed)
        {
            validatedValue = newValue;
        }

        Go::Properties::Value<T>::CheckValidation(newValue);

        if (minimumEnabled && newValue < minimum->Get())
        {
            GoLogUserError(GOPARAMS_VALIDATION_SOURCE, pgettext("sensor", "%1$s was set to be less than its minimum."), 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());

            GoThrowMsg(kERROR_PARAMETER, "%s was set to be less than its minimum.", 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());
        }
        if (maximumEnabled && newValue > maximum->Get())
        {
            GoLogUserError(GOPARAMS_VALIDATION_SOURCE, pgettext("sensor", "%1$s was set to be greater than its maximum."), 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());

            GoThrowMsg(kERROR_PARAMETER, "%s was set to be greater than its maximum.", 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());
        }
        if (exclusiveMinimumEnabled && newValue <= exclusiveMinimum->Get())
        {
            GoLogUserError(GOPARAMS_VALIDATION_SOURCE, pgettext("sensor", "%1$s was set to be less than or equal to its exclusive minimum."), 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());

            GoThrowMsg(kERROR_PARAMETER, "%s was set to be less than or equal to its exclusive minimum.", 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());
        }
        if (exclusiveMaximumEnabled && newValue >= exclusiveMaximum->Get())
        {
            GoLogUserError(GOPARAMS_VALIDATION_SOURCE, pgettext("sensor", "%1$s was set to be greater than or equal to its exclusive maximum."), 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());

            GoThrowMsg(kERROR_PARAMETER, "%s was set to be greater than or equal to its exclusive maximum.", 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());
        }
    }

    DynamicValue<T>& EnableUnits()
    {
        if (!this->units)
        {
            units = new Go::Properties::Value<std::string>();
        }
        this->RegisterSchema("units", *units, false);
        return *this;
    }

    DynamicValue<T>& EnableDescription()
    {
        if (!this->description)
        {
            description = new Go::Properties::Value<std::string>();
        }
        this->RegisterSchema("description", *description, false);
        return *this;
    }

    DynamicValue<T>& EnableReadOnly()
    {
        this->RegisterSchema("readOnly", readOnly, false);
        readOnlyEnabled = true;
        return *this;
    }

    DynamicValue<T>& EnableUsed()
    {
        this->RegisterSchema("used", used, false);
        usedEnabled = true;
        return *this;
    }

    DynamicValue<T>& EnableMinimum()
    {
        if (!this->minimum)
        {
            minimum = new Go::Properties::Value<T>();
        }
        this->RegisterSchema("minimum", *minimum, false);
        minimumEnabled = true;
        return *this;
    }

    DynamicValue<T>& EnableMaximum()
    {
        if (!this->maximum)
        {
            maximum = new Go::Properties::Value<T>();
        }
        this->RegisterSchema("maximum", *maximum, false);
        maximumEnabled = true;
        return *this;
    }

    DynamicValue<T>& EnableExclusiveMinimum()
    {
        if (!this->exclusiveMinimum)
        {
            exclusiveMinimum = new Go::Properties::Value<T>();
        }
        this->RegisterSchema("exclusiveMinimum", *exclusiveMinimum, false);
        exclusiveMinimumEnabled = true;
        return *this;
    }

    DynamicValue<T>& EnableExclusiveMaximum()
    {
        if (!this->exclusiveMaximum)
        {
            exclusiveMaximum = new Go::Properties::Value<T>();
        }
        this->RegisterSchema("exclusiveMaximum", *exclusiveMaximum, false);
        exclusiveMaximumEnabled = true;
        return *this;
    }

    DynamicValue<T>& EnableMinMax()
    {
        return EnableMinimum().EnableMaximum();
    }

    DynamicValue<T>& EnableActual()
    {
        if (!this->actual)
        {
            actual = new Go::Properties::Value<T>();
        }
        this->RegisterSchema("actual", *actual, false);
        actualEnabled = true;
        return *this;
    }

    DynamicValue<T>& EnableScale()
    {
        if (!this->scale)
        {
            scale = new Go::Properties::Value<ScaleOption>();
        }
        this->RegisterSchema("scale", *scale, false);
        return *this;
    }

    const bool MinimumEnabled() const
    {
        return this->minimumEnabled;
    }

    const bool MaximumEnabled() const
    {
        return this->maximumEnabled;
    }

    const bool ExclusiveMinimumEnabled() const
    {
        return this->exclusiveMinimumEnabled;
    }

    const bool ExclusiveMaximumEnabled() const
    {
        return this->exclusiveMaximumEnabled;
    }

    operator const T&() const override
    {
        return Get();
    }

    const T& Get() const override
    {
        return actualEnabled ? actual->Get() : GetUserValue();
    }

    const T& GetUserValue() const
    {
        return Go::Properties::Value<T>::Get();
    }

    void SetExtendedType(const std::string& extendedTypeStr = "")
    {
        if (extendedTypeStr != "")
        {
            if (!this->extendedType)
            {
                extendedType = new Go::Properties::Value<std::string>();
            }
            this->RegisterSchema("extendedType", *extendedType);
            extendedType->Set(extendedTypeStr);
        }
    }

    Go::Properties::Value<std::string>& GetUnits() const
    {
        if (units)
        {
            return *units;
        }
        else
        {
            GoThrowMsg(kERROR_PARAMETER, "Must enable 'units' before using it.");
        }
    }

    Go::Properties::Value<std::string>& GetDescription() const
    {
        if (description)
        {
            return *description;
        }
        else
        {
            GoThrowMsg(kERROR_PARAMETER, "Must enable 'description' before using it.");
        }
    }

    Go::Properties::Value<T>& GetActual() const
    {
        if (actual)
        {
            return *actual;
        }
        else
        {
            GoThrowMsg(kERROR_PARAMETER, "Must enable 'actual' before using it.");
        }
    }

    Go::Properties::Value<T>& GetMinimum() const
    {
        if (minimum)
        {
            return *minimum;
        }
        else
        {
            GoThrowMsg(kERROR_PARAMETER, "Must enable 'minimum' before using it.");
        }
    }

    Go::Properties::Value<T>& GetMaximum()
    {
        if (maximum)
        {
            return *maximum;
        }
        else
        {
            GoThrowMsg(kERROR_PARAMETER, "Must enable 'maximum' before using it.");
        }
    }

    Go::Properties::Value<T>& GetExclusiveMinimum() const
    {
        if (exclusiveMinimum)
        {
            return *exclusiveMinimum;
        }
        else
        {
            GoThrowMsg(kERROR_PARAMETER, "Must enable 'exclusiveMinimum' before using it.");
        }
    }

    Go::Properties::Value<T>& GetExclusiveMaximum() const
    {
        if (exclusiveMaximum)
        {
            return *exclusiveMaximum;
        }
        else
        {
            GoThrowMsg(kERROR_PARAMETER, "Must enable 'exclusiveMaximum' before using it.");
        }
    }

    Go::Properties::Value<ScaleOption>& GetScale() const
    {
        if (scale)
        {
            return *scale;
        }
        else
        {
            GoThrowMsg(kERROR_PARAMETER, "Must enable 'scale' before using it.");
        }
    }

    Go::Properties::Value<std::string>& GetExtendedType() const
    {
        if (extendedType)
        {
            return *extendedType;
        }
        else
        {
            GoThrowMsg(kERROR_PARAMETER, "Must enable 'extendedType' before using it.");
        }
    }

protected:
    bool actualEnabled = false;
    bool minimumEnabled = false;
    bool maximumEnabled = false;
    bool exclusiveMinimumEnabled = false;
    bool exclusiveMaximumEnabled = false;

    void RegisterCommonAttrs()
    {
        this->RegisterSchema("title", title);
        this->RegisterSchema("sortOrder", sortOrder);
    }

private:
    Go::Properties::Value<std::string>* units = nullptr;
    Go::Properties::Value<std::string>* description = nullptr;
    Go::Properties::Value<T>* actual = nullptr;
    Go::Properties::Value<T>* minimum = nullptr;
    Go::Properties::Value<T>* maximum = nullptr;
    // Exclusive min/max don't accept edge values.
    Go::Properties::Value<T>* exclusiveMinimum = nullptr;
    Go::Properties::Value<T>* exclusiveMaximum = nullptr;
    Go::Properties::Value<ScaleOption>* scale = nullptr;
    Go::Properties::Value<std::string>* extendedType = nullptr;
};

}}

#endif // GOAPI_DYNAMIC_PARAMS_H
