#ifndef GOAPI_PARAMS_H
#define GOAPI_PARAMS_H

#include <kVision/L3d/kL3dTransform3d.h>
#include <GoApi/Params/UnsignedValueMin.h>
#include <GoApi/Properties/Nodes.h>

namespace Go {
namespace Params {

#define GOAPI_PARAMS_UNNAMED        "<unnamed>"
#define GOPARAMS_VALIDATION_SOURCE  pgettext("sensor", "Params validation")

#define GOAPI_PARAMS_DEGREE_UNIT                             "deg"      ///< To be consistent with tool side
#define GOAPI_PARAMS_MILLIMETER_UNIT                         "mm"

// Sort order value to ensure it's in position after all the BoxRegion parameters.
#define TRANSFORM_MATRIX_SORT_ORDER_NUMBER 9

// GOS-3627: Possible scales for a GoParams Value.
enum struct ScaleOption : k32u
{
    SCALE_NONE = 1,
    SCALE_LINEAR = 2,
    SCALE_LOGARITHMIC = 3
};

/**
 * Connectedness of points in point set region.
 */
enum struct PointSetRegionMode : k32u
{
    MODE_POINTS = 1,             ///< Individual points
    MODE_OPEN   = 2,             ///< Open, connected in order
    MODE_CLOSED = 3,              ///< Closed loop
    MODE_DEFAULT = k32U_MAX
};

/**
 * Graphic marker shape enumeration.
 */
enum struct MarkerShape : k32s
{
    SHAPE_DEFAULT = -1,
    SHAPE_CROSS = 0,        ///<  Cross marker shape. 
    SHAPE_PLUS = 1,         ///<  Plus marker shape. 
    SHAPE_CIRCLE = 2,       ///<  Circle marker shape. 
    SHAPE_SQUARE = 3,       ///<  Square marker shape. 
    SHAPE_TRIANGLE = 4,     ///<  Triangle marker shape. 
    SHAPE_DIAMOND = 5       ///<  Diamond marker shape. 
};

/**
* A base parameter class for sharing common properties to be inherited by other classes.
*/
struct ParamsBase
{
    Go::Properties::Value<std::string> title;
    Go::Properties::Value<bool> readOnly = false;
    Go::Properties::Value<bool> used = true;
    Go::Properties::Value<k32u> sortOrder;

    // GOS-5307: Ensure readOnly change is only rejected if validation was checked beforehand.
    // This should allow job loading to change readOnly values, as they don't run validation.
    mutable bool readOnlyValidateFailed = false;

    ParamsBase()
    {
    }

    void CheckValidation() const
    {
        if (readOnlyEnabled && readOnly.Get())
        {
            readOnlyValidateFailed = true;
            // GOS-4877: Change readOnly to publish a warning but not fail validation outright.
            // GOS-5061: Only publish this log to the user in debug builds.
            GoLogDebugWarn(GOPARAMS_VALIDATION_SOURCE, "%s is read-only.",
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());
        }
        else
        {
            readOnlyValidateFailed = false;
        }
    }

    void CopyValues(const ParamsBase& other)
    {
        title.Set(other.title);
        sortOrder.Set(other.sortOrder);
        readOnly.Set(other.readOnly);
        used.Set(other.used);

        readOnlyEnabled = other.readOnlyEnabled;
        usedEnabled = other.usedEnabled;
    }

protected:
    bool readOnlyEnabled = false;
    bool usedEnabled = false;
};

/**
 * GoProps values with extra onChange callback functionality to be used for resource update validation.
 * Currently unused. To be finished and utilized in GOS-3163.
 */
template <typename T>
struct ValidationValue : Go::Properties::Value<T>, Event<>
{
    ValidationValue() : 
        Go::Properties::Value<T>()
    {
    }

    ValidationValue(const T& initialValue) : 
        Go::Properties::Value<T>(initialValue)
    {
    }

    ValidationValue& Set(const T& v)
    {
        Go::Properties::Value<T>::Set(v);
        Notify();
        return *this;
    }

    ValidationValue& operator=(const T& v)
    {
        return Set(v);
    }
};

/**
 * A simple "value" parameter class, e.g. a number or a string.
 */
template <typename T>
struct Value : Go::Properties::Value<T>, ParamsBase
{
    Go::Properties::Value<std::string> units;
    Go::Properties::Value<std::string> description;
    Go::Properties::Value<T> actual;
    Go::Properties::Value<T> minimum;
    Go::Properties::Value<T> maximum;
    // Exclusive min/max don't accept edge values.
    Go::Properties::Value<T> exclusiveMinimum;
    Go::Properties::Value<T> exclusiveMaximum;
    Go::Properties::Value<ScaleOption> scale;
    Go::Properties::Value<std::string> extendedType;

    // GOS-5307: Save validated value to ensure read call 
    mutable T validatedValue;

    Value()
    {
        RegisterCommonAttrs();

        T minVal = this->Get();
        bool setMin = UnsignedValueMin<T>::SetUnsignedMin(minVal);
        if (setMin)
        {
            this->EnableMinimum().minimum = minVal;
        }
    }

    Value(const T& val) :
        Go::Properties::Value<T>(val)
    {
        RegisterCommonAttrs();

        T minVal = this->Get();
        bool setMin = UnsignedValueMin<T>::SetUnsignedMin(minVal);
        if (setMin)
        {
            this->EnableMinimum().minimum = minVal;
        }
    }

    // Implementation of virtual method from IValueIo
    void ReadValue(Go::Properties::IValueReader& reader) override
    {
        // GOS-4877: Changed readOnly value doesn't fail validation anymore,
        // but still don't want to change the value.
        // GOS-5307: readOnly shouldn't prevent changes unless validation was previously run.
        if (readOnlyValidateFailed)
        {
            readOnlyValidateFailed = false;
            // If new value matches the one from validation, assume this is the same session.
            // This isn't air-tight, but helps prevent theoretical case where validation sets flag then later fails, 
            // then later a new value is read without validation (such as job load).
            T newValue;
            Go::Properties::ValueIoDispatch<T>::Read(reader, newValue);
            if (Go::Properties::ValueEqFunc<T>::Eq(newValue, validatedValue))
            {
                // Reject change -- readOnly failed validation.
                return;
            }
        }

        Go::Properties::Value<T>::ReadValue(reader);
    }

    Value<T>& operator=(const T& other)
    {
        this->Set(other);
        return *this;
    }

    void CopyValues(const Go::Properties::Node& other)
    {
        Go::Properties::Value<T>::CopyValues(other);

        auto& otherValue = dynamic_cast<const Value<T>&>(other);
        ParamsBase::CopyValues(otherValue);

        if (readOnlyEnabled)
        {
            EnableReadOnly();
        }
        if (usedEnabled)
        {
            EnableUsed();
        }

        units.Set(otherValue.units);
        description.Set(otherValue.description);
        actual.Set(otherValue.actual);
        minimum.Set(otherValue.minimum);
        maximum.Set(otherValue.maximum);
        exclusiveMinimum.Set(otherValue.exclusiveMinimum);
        exclusiveMaximum.Set(otherValue.exclusiveMaximum);
        scale.Set(otherValue.scale);
        extendedType.Set(otherValue.extendedType);

        actualEnabled = otherValue.actualEnabled;
        if (actualEnabled)
        {
            EnableActual();
        }

        minimumEnabled = otherValue.minimumEnabled;
        if (minimumEnabled)
        {
            EnableMinimum();
        }

        maximumEnabled = otherValue.maximumEnabled;
        if (maximumEnabled)
        {
            EnableMaximum();
        }

        exclusiveMinimumEnabled = otherValue.exclusiveMinimumEnabled;
        if (exclusiveMinimumEnabled)
        {
            EnableExclusiveMinimum();
        }

        exclusiveMaximumEnabled = otherValue.exclusiveMaximumEnabled;
        if (exclusiveMaximumEnabled)
        {
            EnableExclusiveMaximum();
        }
    }

    void CheckValidation(const T& newValue) const override
    {
        ParamsBase::CheckValidation();
        if (readOnlyValidateFailed)
        {
            validatedValue = newValue;
        }

        Go::Properties::Value<T>::CheckValidation(newValue);

        if (minimumEnabled && newValue < minimum.Get())
        {
            GoLogUserError(GOPARAMS_VALIDATION_SOURCE, pgettext("sensor", "%1$s was set to be less than its minimum."), 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());

            GoThrowMsg(kERROR_PARAMETER, "%s was set to be less than its minimum.", 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());
        }
        if (maximumEnabled && newValue > maximum.Get())
        {
            GoLogUserError(GOPARAMS_VALIDATION_SOURCE, pgettext("sensor", "%1$s was set to be greater than its maximum."), 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());

            GoThrowMsg(kERROR_PARAMETER, "%s was set to be greater than its maximum.", 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());
        }
        if (exclusiveMinimumEnabled && newValue <= exclusiveMinimum.Get())
        {
            GoLogUserError(GOPARAMS_VALIDATION_SOURCE, pgettext("sensor", "%1$s was set to be less than or equal to its exclusive minimum."), 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());

            GoThrowMsg(kERROR_PARAMETER, "%s was set to be less than or equal to its exclusive minimum.", 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());
        }
        if (exclusiveMaximumEnabled && newValue >= exclusiveMaximum.Get())
        {
            GoLogUserError(GOPARAMS_VALIDATION_SOURCE, pgettext("sensor", "%1$s was set to be greater than or equal to its exclusive maximum."), 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());

            GoThrowMsg(kERROR_PARAMETER, "%s was set to be greater than or equal to its exclusive maximum.", 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());
        }
    }

    Value<T>& EnableUnits()
    {
        this->RegisterSchema("units", units, false);
        return *this;
    }

    Value<T>& EnableDescription()
    {
        this->RegisterSchema("description", description, false);
        return *this;
    }

    Value<T>& EnableReadOnly()
    {
        this->RegisterSchema("readOnly", readOnly, false);
        readOnlyEnabled = true;
        return *this;
    }

    Value<T>& EnableUsed()
    {
        this->RegisterSchema("used", used, false);
        usedEnabled = true;
        return *this;
    }

    Value<T>& EnableMinimum()
    {
        this->RegisterSchema("minimum", minimum, false);
        minimumEnabled = true;
        return *this;
    }

    Value<T>& EnableMaximum()
    {
        this->RegisterSchema("maximum", maximum, false);
        maximumEnabled = true;
        return *this;
    }

    Value<T>& EnableExclusiveMinimum()
    {
        this->RegisterSchema("exclusiveMinimum", exclusiveMinimum, false);
        exclusiveMinimumEnabled = true;
        return *this;
    }

    Value<T>& EnableExclusiveMaximum()
    {
        this->RegisterSchema("exclusiveMaximum", exclusiveMaximum, false);
        exclusiveMaximumEnabled = true;
        return *this;
    }

    Value<T>& EnableMinMax()
    {
        return EnableMinimum().EnableMaximum();
    }

    Value<T>& EnableActual()
    {
        this->RegisterSchema("actual", actual, false);
        actualEnabled = true;
        return *this;
    }

    Value<T>& EnableScale()
    {
        this->RegisterSchema("scale", scale, false);
        return *this;
    }

    const bool MinimumEnabled() const
    {
        return this->minimumEnabled;
    }

    const bool MaximumEnabled() const
    {
        return this->maximumEnabled;
    }

    const bool ExclusiveMinimumEnabled() const
    {
        return this->exclusiveMinimumEnabled;
    }

    const bool ExclusiveMaximumEnabled() const
    {
        return this->exclusiveMaximumEnabled;
    }

    operator const T&() const override
    {
        return Get();
    }

    const T& Get() const override
    {
        return actualEnabled ? actual.Get() : GetUserValue();
    }

    const T& GetUserValue() const
    {
        return Go::Properties::Value<T>::Get();
    }

    void SetExtendedType(const std::string& extendedTypeStr = "")
    {
        if (extendedTypeStr != "")
        {
            this->RegisterSchema("extendedType", extendedType);
            extendedType = extendedTypeStr;
        }
    }

protected:
    bool actualEnabled = false;
    bool minimumEnabled = false;
    bool maximumEnabled = false;
    bool exclusiveMinimumEnabled = false;
    bool exclusiveMaximumEnabled = false;

    void RegisterCommonAttrs()
    {
        this->RegisterSchema("title", title);
        this->RegisterSchema("sortOrder", sortOrder);
    }
};

/**
 * An enum parameter.
 */
template<typename T>
struct Enum : Value<T>
{
    Go::Properties::Value<std::vector<T>> enumVal;
    Go::Properties::Value<std::vector<std::string>> enumText;

    Enum()
    {
        RegisterCommonAttrs();
    }

    Enum(const T& val) : Value<T>(val)
    {
        RegisterCommonAttrs();
    }

    Value<T>& operator=(const T& other)
    {
        this->Set(other);
        return *this;
    }

    void CopyValues(const Go::Properties::Node& other) override
    {
        Value<T>::CopyValues(other);

        auto& otherEnum = dynamic_cast<const Enum<T>&>(other);

        enumVal.Set(otherEnum.enumVal);
        enumText.Set(otherEnum.enumText);
    }

    void CheckValidation(const T& newValue) const
    {
        Value<T>::CheckValidation(newValue);

        if (!HasOption(newValue))
        {
            GoLogUserError(GOPARAMS_VALIDATION_SOURCE, pgettext("sensor", "%1$s was set to a value outside of the enum."), 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());

            GoThrowMsg(kERROR_PARAMETER, "%s was set to a value outside of the enum.", 
                this->title == "" ? GOAPI_PARAMS_UNNAMED : this->title.Get().c_str());
        }
    }

    void ClearOptions()
    {
        enumVal.Set(std::vector<T>());
        enumText.Set(std::vector<std::string>());
    }

    void AddOption(const T& val, const std::string& text)
    {
        auto valVec = enumVal.Get();
        valVec.push_back(val);
        enumVal.Set(valVec);

        auto textVal = enumText.Get();
        textVal.push_back(text);
        enumText.Set(textVal);
    }

    size_t OptionCount() const
    {
        return enumVal.Get().size();
    }

    const T& OptionAt(size_t index) const
    {
        return enumVal.Get().at(index);
    }

    bool HasOption(const T& option) const
    {
        for (auto& val : enumVal.Get())
        {
            if (val == option)
            {
                return true;
            }
        }

        return false;
    }

    /**
     * Returns the registered title of the enumerated parameter.
     *
     * @param   index   Enum value.
     * @return          Reference to the enum text (title).
     */
    const std::string& OptionTextAt(size_t index) const
    {
        return enumText.Get().at(index);
    }

protected:
    void RegisterCommonAttrs()
    {
        this->RegisterSchema("enum", enumVal);
        this->RegisterSchema("enumText", enumText);
    }
};

/**
 * Internal base class for structure-based parameters.
 *
 * This is a base class used internally by the parameter system to implement
 * parameters with multiple values, like regions. It should not be used directly
 * by user code such as tool implementations.
 *
 * A subclass of this class should have its own "extended type", which is used
 * by the client to determine how to render the parameter. Therefore adding
 * a new subclass requires also adding support in the GUI.
 */
struct StructBase : Go::Properties::Structure, ParamsBase
{
    /**
     * The "extended type" of the parameter.
     */
    Go::Properties::Value<std::string> extendedType;

    StructBase& EnableReadOnly()
    {
        this->RegisterSchema("readOnly", readOnly, false);
        readOnlyEnabled = true;
        return *this;
    }

    StructBase& EnableUsed()
    {
        this->RegisterSchema("used", used, false);
        usedEnabled = true;
        return *this;
    }

    void CopyValues(const Go::Properties::Node& other) override
    {
        Go::Properties::Structure::CopyValues(other);

        auto& otherStruct = dynamic_cast<const StructBase&>(other);
        ParamsBase::CopyValues(otherStruct);

        if (readOnlyEnabled)
        {
            EnableReadOnly();
        }
        if (usedEnabled)
        {
            EnableUsed();
        }

        extendedType.Set(otherStruct.extendedType);
        sortOrderCounter = otherStruct.sortOrderCounter;
    }

protected:
    k32u sortOrderCounter;

    /**
     * Constructs an object.
     *
     * @param   extendedTypeStr         The extended type of the parameter.
     */
    StructBase(const std::string& extendedTypeStr = "")
    {
        sortOrderCounter = 0;

        if (extendedTypeStr != "")
        {
            this->RegisterSchema("extendedType", extendedType);
            extendedType = extendedTypeStr;
        }

        this->RegisterSchema("title", title);
        this->RegisterSchema("sortOrder", sortOrder);
    }
};

/**
 * Container for a group of parameters.
 *
 * This class defines a simple container of related parameters. In the GUI a
 * group is rendered as its own section. It is possible to nest groups, but
 * deep nesting is not recommended as it negatively impacts user experience.
 */
struct Group : StructBase
{
    /**
     * By default leave room for 100 sortOrder indexes between two Register calls.
     */
    static constexpr k32s defaultSortOrderStep = 100;

    /**
     * Constructs an object.
     */
    Group(const std::string& extendedTypeStr = "") : StructBase(extendedTypeStr)
    {
    }

    /**
     * Registers a parameter.
     * The order in which a parameter is registered is saved internally with a stepped integer value
     * starting from 0 upon registration within the group. This sort order value if required, can be retrieved
     * through the schema.
     *
     * @param id             Identifies the parameter over the protocol and when persisting values in job files.
     * @param param          Reference to the parameter to register.
     * @param title          The title of the parameter.
     * @param orderingStep   Optional parameter to specify a step for the next sortOrder.
     */
    template <typename PARAM_TYPE>
    void Register(const std::string& id, PARAM_TYPE& param, const std::string& title = "", const k32s orderingStep = defaultSortOrderStep)
    {
        Go::Properties::Structure::Register(id, param);

        param.title = title != "" ? title : id;

        // Save the order of registration.
        param.sortOrder = sortOrderCounter;
        sortOrderCounter += orderingStep;
    }
    void CopyValues(const Go::Properties::Node& other) override
    {
        StructBase::CopyValues(other);
    }

    /**
     * \brief                  Registers a parameter after already registered parameter.
     * \tparam PARAM_TYPE      Parameter type.
     * \param existingParam    Inserts a new parameter after this parameter.
     * \param id               Identifies the parameter over the protocol and when persisting values in job files.
     * \param param            Reference to the parameter to register.
     * \param title            The title of the parameter.
     */
    template <typename PARAM_TYPE>
    void RegisterBefore(const ParamsBase& existingParam, const std::string& id, PARAM_TYPE& param, const std::string& title = "")
    {
        RegisterRelative(existingParam, -1, id, param, title);
    }

    /**
     * \brief                  Registers a parameter before already registered parameter.
     * \tparam PARAM_TYPE      Parameter type.
     * \param existingParam    Inserts a new parameter before this parameter.
     * \param id               Identifies the parameter over the protocol and when persisting values in job files.
     * \param param            Reference to the parameter to register.
     * \param title            The title of the parameter.
     */
    template <typename PARAM_TYPE>
    void RegisterAfter(const ParamsBase& existingParam, const std::string& id, PARAM_TYPE& param, const std::string& title = "")
    {
        RegisterRelative(existingParam, 1, id, param, title);
    }

private:

    template <typename PARAM_TYPE>
    void RegisterRelative(const ParamsBase& existingParam, k32s sortOrderDiff, const std::string& id, PARAM_TYPE& param, const std::string& title = "")
    {
        const auto newSortOrder = existingParam.sortOrder + sortOrderDiff;

        for (auto& [name, node] : children)
        {
            const auto childParam = dynamic_cast<ParamsBase*>(&node);
            if (childParam && childParam->sortOrder == newSortOrder)
            {
                GoThrowMsg(kERROR_ALREADY_EXISTS, "SortOrder %u already registered for %s. Failed to register %s.",
                    newSortOrder, name.c_str(), id.c_str());
            }
        }

        Structure::Register(id, param);

        param.title = !title.empty() ? title : id;

        // Save the order of registration.
        param.sortOrder = newSortOrder;
    }
};

/**
  * An array of objects.
  */
template <typename T>
struct Array : Go::Properties::Array<T>, ParamsBase
{
    Go::Properties::Value<std::string> extendedType;

    Array()
    {
        this->RegisterSchema("title", title);
        this->RegisterSchema("sortOrder", sortOrder);
    }

    template <typename... Args>
    T& Emplace(Args&&... args)
    {
        T& newElement = Go::Properties::Array<T>::Emplace(std::forward<Args>(args)...);
        return newElement;
    }

    void CopyValues(const Go::Properties::Node& other) override
    {
        Go::Properties::Array<T>::CopyValues(other);

        auto& otherArray = dynamic_cast<const Array<T>&>(other);
        ParamsBase::CopyValues(otherArray);

        if (readOnlyEnabled)
        {
            EnableReadOnly();
        }
        if (usedEnabled)
        {
            EnableUsed();
        }

        extendedType.Set(otherArray.extendedType);
    }

    Array& EnableReadOnly()
    {
        this->RegisterSchema("readOnly", readOnly, false);
        readOnlyEnabled = true;
        return *this;
    }

    Array& EnableUsed()
    {
        this->RegisterSchema("used", used, false);
        usedEnabled = true;
        return *this;
    }

    void SetExtendedType(const std::string& extendedTypeStr = "")
    {
        if (extendedTypeStr != "")
        {
            this->RegisterSchema("extendedType", extendedType);
            extendedType = extendedTypeStr;
        }
    }

};

/**
  * The top-level container for parameters.
  */
using ParamSet = Group;



struct PointSetRegionAxisValue : Go::Properties::Value<k64f>
{
    Go::Properties::Value<bool> readOnly;

    PointSetRegionAxisValue()
    {
        RegisterSchema("readOnly", readOnly);
    }

    Value<k64f>& operator=(const k64f& other)
    {
        this->Set(other);
        return *this;
    }
};

/**
 * A point in point set region.
 */
struct PointSetRegionPoint : StructBase
{
    PointSetRegionAxisValue x;
    PointSetRegionAxisValue y;
    PointSetRegionAxisValue z;

    PointSetRegionPoint()
    {
        Register("x", x);
        Register("y", y);
        Register("z", z);
    }
};

struct PointSetRegion : StructBase
{
    Go::Properties::Value<k32u> maxPointCount = k32U_MAX;
    Go::Properties::Value<k32u> minPointCount = 0;
    Go::Properties::Value<MarkerShape> pointShape = MarkerShape::SHAPE_DEFAULT;
    Go::Properties::Value<k16u> pointSize = 0;
    Go::Properties::Value<PointSetRegionMode> mode;
    Go::Properties::Value<PointSetRegionMode> pointColor = PointSetRegionMode::MODE_DEFAULT;
    Go::Properties::Value<PointSetRegionMode> lineColor = PointSetRegionMode::MODE_DEFAULT;
    Go::Properties::Value<bool> showProjection = true;
    Go::Properties::Array<PointSetRegionPoint> points;

    PointSetRegion() :
        constantZEnabled(false),
        constantZValue(0.0)
    {
        RegisterSchema("maxPointCount", maxPointCount);
        RegisterSchema("minPointCount", minPointCount);
        RegisterSchema("pointShape", pointShape);
        RegisterSchema("pointSize", pointSize);
        RegisterSchema("mode", mode);
        RegisterSchema("pointColor", pointColor);
        RegisterSchema("lineColor", lineColor);
        RegisterSchema("showProjection", showProjection);
        
        Register("point", points);
    }

    /**
      * Copy values from one Node to this Node.
      *
      * @param other         Node to copy values from.
      */
    void CopyValues(const Node& other) override
    {
        auto& typedOther = dynamic_cast<const PointSetRegion&>(other);

        maxPointCount = typedOther.maxPointCount.Get();
        minPointCount = typedOther.minPointCount.Get();
        pointShape = typedOther.pointShape.Get();
        pointSize = typedOther.pointSize.Get();
        mode = typedOther.mode.Get();
        pointColor = typedOther.pointColor.Get();
        lineColor = typedOther.lineColor.Get();
        showProjection = typedOther.showProjection.Get();

        for (size_t i = 0; i < typedOther.points.Count(); ++i)
        {
            points.At(i).x.Set(typedOther.points.At(i).x.Get());
            points.At(i).x.readOnly.Set(typedOther.points.At(i).x.readOnly);

            points.At(i).y.Set(typedOther.points.At(i).y.Get());
            points.At(i).y.readOnly.Set(typedOther.points.At(i).y.readOnly);

            points.At(i).z.Set(typedOther.points.At(i).z.Get());
            points.At(i).z.readOnly.Set(typedOther.points.At(i).z.readOnly);
        }
    }

    /**
     * Enable and set the constant z value.
     *
     * @param zValue         z value to copy.
     */
    void EnableConstantZ(k64f zValue)
    {
        constantZEnabled = true;
        constantZValue = zValue;
        UpdateConstantZ();
    }
    
    /**
     * Disable the constant z value.
     */
    void DisableConstantZ()
    {
        constantZEnabled = false;
        UpdateConstantZ();
    }
    
    /**
     * Set the point count value.
     *
     * @param count         count value to copy.
     */
    void SetPointCount(size_t count)
    {
        points.Resize(count);
        UpdateConstantZ();
    }

protected:
   
    bool constantZEnabled;
    k64f constantZValue;
    
    /**
     * Update the constant z values.
     */
    void UpdateConstantZ()
    {
        for (size_t i = 0; i < points.Count(); ++i)
        {
            points.At(i).z.readOnly = constantZEnabled;

            if (constantZEnabled)
            {
                points.At(i).z.Set(constantZValue);
            }
        }
    }
};


// Regions extended types.
/**
 * A full 3D box region parameter with optional angle support.
 * All parameters are optional although the respective coordinate and region dimension 
 * are paired up to minimize the number of parameters in the constructor.
 * Boolean flags passed to the constructor determines which parameters are used/registered.
 * The default extended type is "TransformedBoxRegion".
 *
 * The "sortOrder" attributes were utilized here to ensure the region parameters appear on
 * the client side in expected order, namely:
 *    X
 *    Y
 *    Z
 *    Width  (X)
 *    Length (Y)
 *    Height (Z)
 *    X Angle
 *    Y Angle
 *    Z Angle
 */
struct BoxRegion : Group
{
    Go::Params::Value<k64f> x;
    Go::Params::Value<k64f> y;
    Go::Params::Value<k64f> z;
    Go::Params::Value<k64f> width;
    Go::Params::Value<k64f> length;
    Go::Params::Value<k64f> height;
    Go::Params::Value<k64f> xAngle;
    Go::Params::Value<k64f> yAngle;
    Go::Params::Value<k64f> zAngle;

    BoxRegion(
        bool useXWidth = true, 
        bool useYLength = true, 
        bool useZHeight = true,
        bool useXAngle = false, 
        bool useYAngle = false, 
        bool useZAngle = false,
        std::string extendedTypeStr = "TransformedBoxRegion") : 
        
        Group(extendedTypeStr)
    {
        if (useXWidth)
        {
            Register("x", x, "X");
            Register("width", width, pgettext("sensor", "Width (X)"));
            x.sortOrder.Set(0);
            width.sortOrder.Set(3);
            x.EnableUnits().units.Set(GOAPI_PARAMS_MILLIMETER_UNIT);
            width.EnableUnits().units.Set(GOAPI_PARAMS_MILLIMETER_UNIT);
        }
        else
        {
            x.EnableUsed();
            x.used = false;
            width.EnableUsed();
            width.used = false;
        }
        if (useYLength)
        {
            Register("y", y, "Y");
            Register("length", length, pgettext("sensor", "Length (Y)"));
            y.sortOrder.Set(1);
            length.sortOrder.Set(4);
            y.EnableUnits().units.Set(GOAPI_PARAMS_MILLIMETER_UNIT);
            length.EnableUnits().units.Set(GOAPI_PARAMS_MILLIMETER_UNIT);
        }
        else
        {
            y.EnableUsed();
            y.used = false;
            length.EnableUsed();
            length.used = false;
        }
        if (useZHeight)
        {
            Register("z", z, "Z");
            Register("height", height, pgettext("sensor", "Height (Z)"));
            z.sortOrder.Set(2);
            height.sortOrder.Set(5);
            z.EnableUnits().units.Set(GOAPI_PARAMS_MILLIMETER_UNIT);
            height.EnableUnits().units.Set(GOAPI_PARAMS_MILLIMETER_UNIT);
        }
        else
        {
            z.EnableUsed();
            z.used = false;
            height.EnableUsed();
            height.used = false;
        }

        if (useXAngle)
        {
            Register("xAngle", xAngle, pgettext("sensor", "X Angle"));
            xAngle.sortOrder.Set(6);
            xAngle.EnableUnits().units.Set(GOAPI_PARAMS_DEGREE_UNIT);
        }
        else
        {
            xAngle.EnableUsed();
            xAngle.used = false;
        }
        if (useYAngle)
        {
            Register("yAngle", yAngle, pgettext("sensor", "Y Angle"));
            yAngle.sortOrder.Set(7);
            yAngle.EnableUnits().units.Set(GOAPI_PARAMS_DEGREE_UNIT);
        }
        else
        {
            yAngle.EnableUsed();
            yAngle.used = false;
        }
        if (useZAngle)
        {
            Register("zAngle", zAngle, pgettext("sensor", "Z Angle"));
            zAngle.sortOrder.Set(8);
            zAngle.EnableUnits().units.Set(GOAPI_PARAMS_DEGREE_UNIT);
        }
        else
        {
            zAngle.EnableUsed();
            zAngle.used = false;
        }
    }

    /**
     * Set this structure's values with values passed.
     *
     * @param xVal          x value.
     * @param yVal          y value.
     * @param zVal          z value.
     * @param widthVal      width value.
     * @param lengthVal     length value.
     * @param heightVal     height value.
     */
    void Set(k64f xVal, k64f yVal, k64f zVal, k64f widthVal, k64f lengthVal, k64f heightVal)
    {
        x = xVal;
        y = yVal;
        z = zVal;
        width = widthVal;
        length = lengthVal;
        height = heightVal;
    }

    /**
     * Set this structure's angle values with values passed.
     *
     * @param xAngleVal     x angle value.
     * @param yAngleVal     y angle value.
     * @param zAngleVal     z angle value.
     */
    void SetAngles(k64f xAngleVal, k64f yAngleVal, k64f zAngleVal)
    {
        xAngle = xAngleVal;
        yAngle = yAngleVal;
        zAngle = zAngleVal;
    }

    /**
     * Converts this structure's values to a kRect3d64f structure object.
     * NOTE:
     *  - kRect3d64f's height field corresponds to Gocator length field along the Y-axis.
     *  - kRect3d64f's depth field corresponds to Gocator height field along the Z-axis.
     *
     * @return  The converted structure.
     */
    kRect3d64f ToRect3d() const
    {
        kRect3d64f rect3d;

        rect3d.x = x;
        rect3d.y = y;
        rect3d.z = z;
        rect3d.width = width;
        rect3d.height = length;
        rect3d.depth = height;

        return rect3d;
    }

    /**
     * Update this structure's values from the kRect3d64f structure object passed.
     * NOTE:
     *  - kRect3d64f's height field correspnds to Gocator length field along the Y-axis.
     *  - kRect3d64f's depth field correspnds to Gocator height field along the Z-axis.
     *
     * @param rect3d         kRect3d64f structure object.
     */
    void FromRect3d(const kRect3d64f& rect3d)
    {
        x = rect3d.x;
        y = rect3d.y;
        z = rect3d.z;
        width = rect3d.width;
        length = rect3d.height;
        height = rect3d.depth;
    }
};

/**
 * A subset of the 3D BoxRegion only enabling parameters for a 2D profile region,
 * namely x, z, width, height and yAngle. yAngle is disabled by default.
 */
struct ProfileRegion : BoxRegion
{
    ProfileRegion(bool includeYRotation = false) :
        BoxRegion(true, false, true, false, includeYRotation, false)
    {
    }

    /**
     * Set profile region's values with values passed.
     *
     * @param xVal          x value.
     * @param zVal          z value.
     * @param widthVal      width value.
     * @param heightVal     height value.
     * @param yAngleVal     y angle value.
     */
    void Set(k64f xVal, k64f zVal, k64f widthVal, k64f heightVal, k64f yAngleVal = 0.0)
    {
        x = xVal;
        z = zVal;
        width = widthVal;
        height = heightVal;
        yAngle = yAngleVal;
    }
};

/**
 * A subset of the 3D BoxRegion only enabling parameters for a 2D surface region.
 * namely x, y, width, length and zAngle. zAngle is disabled by default.
 *
 * @param includeZRotation      Set to true if want to use Z angle, otherwise set to false.
 *                              Default to not using Z angle so give a purely 2-D region.
*/
struct SurfaceRegion2d : BoxRegion
{
    SurfaceRegion2d(bool includeZRotation = false) :
        BoxRegion(true, true, false, false, false, includeZRotation)
    {
    }

    /**
     * Set surface region's values with values passed.
     *
     * @param xVal          x value.
     * @param yVal          y value.
     * @param widthVal      width value.
     * @param lengthVal     length value.
     * @param zAngleVal     z angle value.
     */
    void Set(
        k64f xVal, 
        k64f yVal, 
        k64f widthVal, 
        k64f lengthVal, 
        k64f zAngleVal = 0.0)
    {
        x = xVal;
        y = yVal;
        width = widthVal;
        length = lengthVal;
        zAngle = zAngleVal;
    }
};

/**
 * GOS-3029: A subset of the 3D BoxRegion only enabling parameters for a 3D surface region but
 * with only Z angle (ie. no X angle or Y angle). 
 * In other words, this surface region provides X/Y/Z/width/length/height/Zangle.
 * This is what VE team wants.
 *
 * @param includeZRotation      Set to true if want to use Z angle, otherwise set to false.
 *                              Default (arbitrarily) to not use Z angle.
*/
struct SurfaceRegion3d : BoxRegion
{
    SurfaceRegion3d(bool includeZRotation = false) :
        BoxRegion(true, true, true, false, false, includeZRotation)
    {}

    /**
     * Set surface region's values with values passed.
     *
     * @param xVal          x value.
     * @param yVal          y value.
     * @param zVal          z value.
     * @param widthVal      width value.
     * @param lengthVal     length value.
     * @param heightVal     height value.
     * @param zAngleVal     z angle value.
     */
    void Set(
        k64f xVal, 
        k64f yVal, 
        k64f zVal, 
        k64f widthVal, 
        k64f lengthVal, 
        k64f heightVal, 
        k64f zAngleVal = 0.0)
    {
        x = xVal;
        y = yVal;
        z = zVal;
        width = widthVal;
        length = lengthVal;
        height = heightVal;
        zAngle = zAngleVal;
    }
};

// 3D Transform matrix structure.
/**
 * A 3D transformation matrix parameter.
        [ xx xy xz xt ]
        [ yx yy yz yt ]
        [ zx zy zz zt ]
        [ 0  0  0  1  ]

 * This matrix is represented as 16 element array in the following order:
 * xx xy xz xt yx yy yx yt zx zy zz zt 0 0 0 1
 * An empty array will be used to indicate an invalid transform.
 */
struct Transform3d : Go::Params::Value<std::vector<k64f>>
{
    Transform3d()
    {
        this->EnableReadOnly();
        this->readOnly = true;
        this->SetExtendedType("Transform3d");
        Reset();
    }

    /**
     * Reset this transform's values to values corresponding to 
     * 4 x 4 identity matrix. This by definition makes the transform valid.
     * 
     *  
     */
    void Reset()
    {
        std::vector<k64f> matrixVector = { 1.0, 0.0, 0.0, 0.0,
                                           0.0, 1.0, 0.0, 0.0,
                                           0.0, 0.0, 1.0, 0.0,
                                           0.0, 0.0, 0.0, 1.0 };

        Set(matrixVector);
    }

    void SetAsInvalid()
    {
        std::vector<k64f> matrixVector;
        Set(matrixVector);
    }

    bool IsValid()
    {
        return !Get().empty();
    }

    /**
     * Returns a kL3dTransform3d transform with values copied from this
     * structure's transform object. As the last row is irrelevant to
     * the kL3dTransform3d structure, it is not copied to this structure.
     *
     * @return  The kL3dTransform3d tranform.
     */
    kL3dTransform3d ToTransform()
    {
        kL3dTransform3d transformOut;

        if (!Get().empty())
        {
            transformOut.xx = Get()[0];
            transformOut.xy = Get()[1];
            transformOut.xz = Get()[2];
            transformOut.xt = Get()[3];
            transformOut.yx = Get()[4];
            transformOut.yy = Get()[5];
            transformOut.yz = Get()[6];
            transformOut.yt = Get()[7];
            transformOut.zx = Get()[8];
            transformOut.zy = Get()[9];
            transformOut.zz = Get()[10];
            transformOut.zt = Get()[11];
        }
        else
        {
            // Default to identity matrix.
            transformOut.xx = 1.0;
            transformOut.xy = 0.0;
            transformOut.xz = 0.0;
            transformOut.xt = 0.0;
            transformOut.yx = 0.0;
            transformOut.yy = 1.0;
            transformOut.yz = 0.0;
            transformOut.yt = 0.0;
            transformOut.zx = 0.0;
            transformOut.zy = 0.0;
            transformOut.zz = 1.0;
            transformOut.zt = 0.0;
        }

        return transformOut;
    }

    /**
     * Set this structure's transform object values with values in
     * the passed-in transform. As the kL3dTransform3d structure
     * does not contain values for the last row of the 4x4 matrix,
     * default data is inserted here.
     *
     * @param transformIn   transform values to copy from
     */
    void FromTransform(const kL3dTransform3d& transformIn)
    {
        std::vector<k64f> matrixVector;

        matrixVector.push_back(transformIn.xx);
        matrixVector.push_back(transformIn.xy);
        matrixVector.push_back(transformIn.xz);
        matrixVector.push_back(transformIn.xt);
        matrixVector.push_back(transformIn.yx);
        matrixVector.push_back(transformIn.yy);
        matrixVector.push_back(transformIn.yz);
        matrixVector.push_back(transformIn.yt);
        matrixVector.push_back(transformIn.zx);
        matrixVector.push_back(transformIn.zy);
        matrixVector.push_back(transformIn.zz);
        matrixVector.push_back(transformIn.zt);
        matrixVector.push_back(0.0);
        matrixVector.push_back(0.0);
        matrixVector.push_back(0.0);
        matrixVector.push_back(1.0);

        Set(matrixVector);
    }
};


// Transformed regions derived types.

/**
 * A full 3D box region with transform and optional angle support.
 * Individual parameters can be individually configured but the default
 * is having all of them enabled.
 * Parameters are x, y, z, width, length, height, xAngle, yAngle, zAngle, transform.
 */
struct TransformedBoxRegion : BoxRegion
{
    Transform3d transformMatrix;

    TransformedBoxRegion(
        bool useXWidth = true, 
        bool useYLength = true, 
        bool useZHeight = true,
        bool useXAngle = true, 
        bool useYAngle = true, 
        bool useZAngle = true) :

        BoxRegion(useXWidth, useYLength, useZHeight, useXAngle, useYAngle, useZAngle)
    {
        Register("transformMatrix", transformMatrix, pgettext("sensor", "Transform Matrix"));
        // Set the sort order of the transformMatrix to be sorted after the BoxRegion parameters.
        transformMatrix.sortOrder.Set(TRANSFORM_MATRIX_SORT_ORDER_NUMBER);
    }
};

/**
 * A subset of the full 3D TransformedBoxRegion only enabling properties for a 
 * 2D transformed profile region with optional angle support.
 * Parameters are x, z, width, height, yAngle, transform. yAngle is optional but 
 * enabled by default for this region.
 */
struct TransformedProfileRegion : ProfileRegion
{
    Transform3d transformMatrix;

    TransformedProfileRegion(bool includeYRotation = true) :
        ProfileRegion(includeYRotation)
    {
        Register("transformMatrix", transformMatrix, pgettext("sensor", "Transform Matrix"));
        // Set the sort order of the transformMatrix to be sorted after the BoxRegion parameters.
        transformMatrix.sortOrder.Set(TRANSFORM_MATRIX_SORT_ORDER_NUMBER);
    }
};

/**
 * A subset of the full 3D TransformedBoxRegion only enabling properties for a
 * 2D transformed surface region with optional angle support.
 * Parameters are x, y, width, height, zAngle tranform. zAngle is optional but
 * enabled by default for this region.
*/
struct TransformedSurfaceRegion : SurfaceRegion2d
{
    Transform3d transformMatrix;

    TransformedSurfaceRegion(bool includeZRotation = true) :
        SurfaceRegion2d(includeZRotation)
    {
        Register("transformMatrix", transformMatrix, pgettext("sensor", "Transform Matrix"));
        // Set the sort order of the transformMatrix to be sorted after the BoxRegion parameters.
        transformMatrix.sortOrder.Set(TRANSFORM_MATRIX_SORT_ORDER_NUMBER);
    }
};

}}

#endif
