/**
* @file    TypeInfo.h
* @brief   Defines type traits used by the properties system.
*
*/

#ifndef GOAPI_PROPERTIES_TYPE_INFO_H
#define GOAPI_PROPERTIES_TYPE_INFO_H

#include <GoApi/GoApiDef.h>
#include <kApi/Data/kArrayList.h>
#include <array>
#include <vector>
#include <deque>
#include <unordered_set>
#include <list>
#include <string>
#include <array>
#include <set>
#include <map>
#include <unordered_map>
#include <forward_list>

namespace Go
{
namespace Properties
{

// Type to define a binary type. Serializers will serialize this into binary type. std::vector<kByte> still serializes into an array.
// std::vector cannot be type-cast to ByteVector; the copy constructor can be used instead.
// ByteVector is otherwise identical to std::vector<kByte>.
struct ByteVector : std::vector<kByte>
{
    using std::vector<kByte>::vector;

    ByteVector() : std::vector<kByte>()
    {
    }

    ByteVector(std::vector<kByte> other) : std::vector<kByte>(other)
    {
    }

    // GcBase64_Decode decodes into an ArrayList. This helps convert that to a ByteVector.
    ByteVector(kArrayList bytes)
        : std::vector<kByte>(
            (kByte*)kArrayList_At(bytes, 0),
            (kByte*)kArrayList_At(bytes, 0) + kArrayList_Count(bytes)
        )
    {
    }
};

namespace Internal
{

// IsTypeArray
template<typename T, typename = void>
struct IsTypeArray
{
    static constexpr bool value = false;
};

template<typename T, size_t SIZE>
struct IsTypeArray<std::array<T, SIZE>>
{
    static constexpr bool value = true;
};

// There are more generic ways of doing this. But this should suffice for our purposes.
#define DECLARE_TYPE_AS_ARRAY(TYPE)             \
template<typename T>                            \
struct IsTypeArray<TYPE<T>>                     \
{                                               \
    static constexpr bool value = true;         \
};

DECLARE_TYPE_AS_ARRAY(std::vector)
DECLARE_TYPE_AS_ARRAY(std::deque)
DECLARE_TYPE_AS_ARRAY(std::forward_list)
DECLARE_TYPE_AS_ARRAY(std::list)
DECLARE_TYPE_AS_ARRAY(std::set)
DECLARE_TYPE_AS_ARRAY(std::multiset)
DECLARE_TYPE_AS_ARRAY(std::unordered_multiset)
DECLARE_TYPE_AS_ARRAY(std::unordered_set)

// IsTypeCArray
template<typename T, typename = void>
struct IsTypeCArray
{
    static constexpr bool value = false;
};

template<typename T, size_t SIZE>
struct IsTypeCArray<std::array<T, SIZE>>
{
    static constexpr bool value = true;
};

// IsTypeObject
template<typename T, typename = void>
struct IsTypeObject
{
    static constexpr bool value = false;
};

#define DECLARE_TYPE_AS_OBJECT(TYPE)            \
template<typename... T>                         \
struct IsTypeObject<TYPE<T...>>                 \
{                                               \
    static constexpr bool value = true;         \
};

DECLARE_TYPE_AS_OBJECT(std::map)
DECLARE_TYPE_AS_OBJECT(std::unordered_map)
DECLARE_TYPE_AS_OBJECT(std::multimap)
DECLARE_TYPE_AS_OBJECT(std::unordered_multimap)

// TypeValueName
template<typename T, typename = void>
struct TypeValueName
{
    static constexpr const char* value = "unknown";
};

template<typename T>
struct TypeValueName<T, std::enable_if_t<std::is_integral<T>::value && !std::is_same<T, bool>::value>>
{
    static constexpr const char* value = "integer";
};

template<typename T>
struct TypeValueName<T, std::enable_if_t<std::is_enum<T>::value>>
{
    static constexpr const char* value = "integer";
};

template<typename T>
struct TypeValueName<T, std::enable_if_t<std::is_floating_point<T>::value>>
{
    static constexpr const char* value = "number";
};

template<typename T>
struct TypeValueName<T, std::enable_if_t<std::is_same<T, bool>::value>>
{
    static constexpr const char* value = "boolean";
};

template<typename T>
struct TypeValueName<T, std::enable_if_t<std::is_same<T, std::string>::value>>
{
    static constexpr const char* value = "string";
};

template<typename T>
struct TypeValueName<T, std::enable_if_t<std::is_same<T, ByteVector>::value>>
{
    static constexpr const char* value = "binary";
};

template<typename T>
struct TypeValueName<T, std::enable_if_t<IsTypeArray<T>::value>>
{
    static constexpr const char* value = "array";
};

template<typename T>
struct TypeValueName<T, std::enable_if_t<IsTypeObject<T>::value>>
{
    static constexpr const char* value = "object";
};

// ArrayItemType
template<typename T, typename = void>
struct ArrayItemType
{
    using type = void;
};

template<typename T>
struct ArrayItemType<T, std::enable_if_t<IsTypeArray<T>::value>>
{
    using type = typename T::value_type;
};

}}} //Namespaces

#endif
