#include <GoApi/Properties/Serializers/JsonSerializer.h>
#include <GoApi/Properties/Nodes/TypeInfo.h>

using nlohmann::json;

namespace Go
{
namespace Properties
{

class JsonValueWriter : public IValueWriter
{
public:
    JsonValueWriter() :
        isWritingArray(false)
    {
    }

    ~JsonValueWriter()
    {
    }

    // Sets the JSON value to write to
    void SetJsonValue(nlohmann::json::pointer jsonValue)
    {
        this->jsonValue = jsonValue;
    }

    void WriteChar(k8s value) override
    {
        if (value == k8S_NULL)
        {
            SetPrimitiveValue(nullptr);
        }
        else
        {
            SetPrimitiveValue(value);
        }
    }

    void WriteUChar(k8u value) override
    {
        if (value == k8U_NULL)
        {
            SetPrimitiveValue(nullptr);
        }
        else
        {
            SetPrimitiveValue(value);
        }
    }

    void WriteShort(k16s value) override
    {
        if (value == k16S_NULL)
        {
            SetPrimitiveValue(nullptr);
        }
        else
        {
            SetPrimitiveValue(value);
        }
    }

    void WriteUShort(k16u value) override
    {
        if (value == k16U_NULL)
        {
            SetPrimitiveValue(nullptr);
        }
        else
        {
            SetPrimitiveValue(value);
        }
    }


    void WriteInt(k32s value) override
    {
        if (value == k32S_NULL)
        {
            SetPrimitiveValue(nullptr);
        }
        else
        {
            SetPrimitiveValue(value);
        }
    }

    void WriteUInt(k32u value) override
    {
        if (value == k32U_NULL)
        {
            SetPrimitiveValue(nullptr);
        }
        else
        {
            SetPrimitiveValue(value);
        }
    }

    void WriteLong(k64s value) override
    {
        if (value == k64S_NULL)
        {
            SetPrimitiveValue(nullptr);
        }
        else
        {
            SetPrimitiveValue(value);
        }
    }

    void WriteULong(k64u value) override
    {
        if (value == k64U_NULL)
        {
            SetPrimitiveValue(nullptr);
        }
        else
        {
            SetPrimitiveValue(value);
        }
    }

    void WriteFloat(k32f value) override
    {
        if (value == k32F_NULL)
        {
            SetPrimitiveValue(nullptr);
        }
        else
        {
            SetPrimitiveValue(value);
        }
    }

    void WriteDouble(k64f value) override
    {
        if (value == k64F_NULL)
        {
            SetPrimitiveValue(nullptr);
        }
        else
        {
            SetPrimitiveValue(value);
        }
    }

    void WriteBool(bool value) override
    {
        SetPrimitiveValue(value);
    }

    void WriteString(const kChar* str) override
    {
        if (isWritingArray)
        {
            jsonValue->push_back(std::string(str));
        }
        else
        {
            *jsonValue = std::string(str);
        }
    }

    void WriteBinary(const ByteVector bin) override
    {
        auto binaryJson = nlohmann::json::binary(bin);

        if (isWritingArray)
        {
            jsonValue->push_back(binaryJson);
        }
        else
        {
            *jsonValue = binaryJson;
        }
    }

    void BeginWriteArray() override
    {
        *jsonValue = nlohmann::json::array();
        isWritingArray = true;
    }

    void EndWriteArray() override
    {
        isWritingArray = false;
    }

private:
    nlohmann::json::pointer jsonValue;
    bool isWritingArray;

    template <typename T>
    void SetPrimitiveValue(T value)
    {
        if (isWritingArray)
        {
            jsonValue->push_back(value);
        }
        else
        {
            *jsonValue = value;
        }
    }
};

class JsonValueReader : public IValueReader
{
public:
    JsonValueReader() :
        arrayState(ArrayState::NONE)
    {
    }

    ~JsonValueReader()
    {
    }

    // Sets the JSON value to read from
    void SetJsonValue(nlohmann::json::const_pointer jsonValue)
    {
        this->jsonValue = jsonValue;
    }

    k8s ReadChar() override
    {
        if (ValueIsNull())
        {
            return k8S_NULL;
        }
        return ReadValue<k8s>();
    }

    k8u ReadUChar() override
    {
        if (ValueIsNull())
        {
            return k8U_NULL;
        }

        if (ValueIsNegative<k8u>())
        {
            GoThrowMsg(kERROR_PARAMETER, "Expected unsigned integer but the value is negative.");
        }
        
        return ReadValue<k8u>();
    }

    k16s ReadShort() override
    {
        if (ValueIsNull())
        {
            return k16S_NULL;
        }
        return ReadValue<k16s>();
    }

    k16u ReadUShort() override
    {
        if (ValueIsNull())
        {
            return k16U_NULL;
        }

        if (ValueIsNegative<k16u>())
        {
            GoThrowMsg(kERROR_PARAMETER, "Expected unsigned integer but the value is negative.");
        }

        return ReadValue<k16u>();
    }

    k32s ReadInt() override
    {
        if (ValueIsNull())
        {
            return k32S_NULL;
        }
        return ReadValue<k32s>();
    }

    k32u ReadUInt() override
    {
        if (ValueIsNull())
        {
            return k32U_NULL;
        }

        if (ValueIsNegative<k32u>())
        {
            GoThrowMsg(kERROR_PARAMETER, "Expected unsigned integer but the value is negative.");
        }

        return ReadValue<k32u>();
    }

    k64s ReadLong() override
    {
        if (ValueIsNull())
        {
            return k64S_NULL;
        }
        return ReadValue<k64s>();
    }

    k64u ReadULong() override
    {
        if (ValueIsNull())
        {
            return k64U_NULL;
        }

        if (ValueIsNegative<k64u>())
        {
            GoThrowMsg(kERROR_PARAMETER, "Expected unsigned integer but the value is negative.");
        }

        return ReadValue<k64u>();
    }

    k32f ReadFloat() override
    {
        if (ValueIsNull())
        {
            return k32F_NULL;
        }
        return ReadValue<k32f>();
    }

    k64f ReadDouble() override
    {
        if (ValueIsNull())
        {
            return k64F_NULL;
        }
        return ReadValue<k64f>();
    }

    bool ReadBool() override
    {
        return ReadValue<bool>();
    }

    std::string ReadString() override
    {
        return ReadValue<std::string>();
    }

    ByteVector ReadBinary() override
    {
        JsonSerializer serializer;
        if (arrayState == ArrayState::READING)
        {
            return serializer.DecipherBinaryJson((*jsonValue)[arrayIndex]);
        }
        else
        {
            return serializer.DecipherBinaryJson(*jsonValue);
        }
    }

    void BeginReadArray() override
    {
        if (jsonValue->is_array())
        {
            arrayState = ArrayState::INIT;
        }
        else
        {
            arrayState = ArrayState::NONE;
        }
    }

    bool NextArrayItem() override
    {
        switch (arrayState)
        {
        case ArrayState::INIT:
            arrayIndex = 0;
            break;
        case ArrayState::READING:
            arrayIndex++;
            break;
        case ArrayState::NONE:
            return false;
        }

        if (arrayIndex < jsonValue->size())
        {
            arrayState = ArrayState::READING;
        }
        else
        {
            arrayState = ArrayState::NONE;
        }

        return arrayState == ArrayState::READING;
    }

    void EndReadArray() override
    {
        arrayState = ArrayState::NONE;
    }

private:
    nlohmann::json::const_pointer jsonValue;

    enum struct ArrayState
    {
        NONE,
        INIT,
        READING
    };

    ArrayState arrayState;
    size_t arrayIndex;

    template <typename T>
    T ReadValue()
    {
        if (arrayState == ArrayState::READING)
        {
            return (*jsonValue)[arrayIndex].get<T>();
        }
        else
        {
            return jsonValue->get<T>();
        }
    }

    bool ValueIsNull()
    {
        if (arrayState == ArrayState::READING)
        {
            return (*jsonValue)[arrayIndex].is_null();
        }
        else
        {
            return jsonValue->is_null();
        }
    }

    template <typename T>
    bool ValueIsNegative()
    {
        if (!std::is_arithmetic<T>())
        {
            // Non-numeric (int/float) types are never negative.
            return false;
        }

        // 0 is signed by default, which makes the json value signed too when compared.
        // This makes large unsigned json values negative, leading to false positives.
        // Casting 0 unsigned prevents this.
        if (arrayState == ArrayState::READING)
        {
            return (*jsonValue)[arrayIndex] < (k8u)0;
        }
        else
        {
            return (*jsonValue) < (k8u)0;
        }
    }
};

class JsonSerializerImpl : public INodeVisitorConst
{
public:
    nlohmann::json Serialize(const Node& node)
    {
        nlohmann::json output;

        stack.emplace(&node, &output);

        while (!stack.empty())
        {
            const Node& curNode = *stack.top().node;
            curJson = stack.top().jsonPtr;
            stack.pop();

            curNode.Accept(*this);
        }

        return output;
    }

    void Visit(const Structure& node) override
    {
        *curJson = json::object();

        for (auto it = node.begin(); it != node.end(); it++)
        {
            auto& jsonItem = (*curJson)[it->first];
            stack.emplace(&it->second, &jsonItem);
        }
    }

    void Visit(const Internal::ArrayBase& node) override
    {
        *curJson = json(node.Count(), nullptr);

        for (size_t i = 0; i < node.Count(); ++i)
        {
            auto& jsonItem = (*curJson)[i];
            stack.emplace(node.At(i), &jsonItem);
        }
    }

    void Visit(const Internal::ReferenceArray& node) override
    {
        *curJson = json(node.Count(), nullptr);

        for (size_t i = 0; i < node.Count(); ++i)
        {
            auto& jsonItem = (*curJson)[i];
            stack.emplace(node.At(i), &jsonItem);
        }
    }

    void Visit(const Internal::ValueBase& node) override
    {
        valueWriter.SetJsonValue(curJson);
        node.WriteValue(valueWriter);
    }

private:
    struct NodeRef
    {
        const Node* node;
        nlohmann::json::pointer jsonPtr;

        NodeRef(const Node* node, nlohmann::json::pointer jsonPtr) :
            node(node), jsonPtr(jsonPtr)
        {
        }
    };

    std::stack<NodeRef> stack;
    nlohmann::json::pointer curJson;
    JsonValueWriter valueWriter;
};

class JsonDeserializerImpl : public INodeVisitor
{
public:
    void Deserialize(Node& node, const nlohmann::json& json)
    {
        stack.emplace(&node, &json);

        while (!stack.empty())
        {
            Node& curNode = *stack.top().node;
            curJson = stack.top().jsonPtr;
            stack.pop();

            curNode.Accept(*this);
        }
    }

    void Visit(Structure& node) override
    {
        node.SetIsRead(curJson != nullptr);

        for (auto it = node.begin(); it != node.end(); ++it)
        {
            auto& name = it->first;
            auto& node = it->second;

            if (node.WriteOnly())
            {
                continue;
            }

            json::const_pointer jsonItem = nullptr;

            if (curJson)
            {
                auto jsonIt = curJson->find(name);
                if (jsonIt != curJson->end())
                {
                    jsonItem = &jsonIt.value();
                }
            }

            stack.emplace(&node, jsonItem);
        }
    }

    void Visit(Internal::ArrayBase& node) override
    {
        node.SetIsRead(curJson != nullptr);

        if (curJson)
        {
            // Modified based on GOS-1337
            GoThrowIf(!curJson->is_array(), kERROR_PARAMETER);

            node.Resize(curJson->size());

            for (size_t i = 0; i < node.Count(); ++i)
            {
                stack.emplace(node.At(i), &curJson->at(i));
            }
        }
        else
        {
            for (size_t i = 0; i < node.Count(); ++i)
            {
                stack.emplace(node.At(i), nullptr);
            }
        }        
    }

    void Visit(Internal::ReferenceArray& node) override
    {
        // ReferenceArray is used only in schemas, which are never read.
    }

    void Visit(Internal::ValueBase& node) override
    {
        node.SetIsRead(curJson != nullptr);

        if (curJson)
        {
            valueReader.SetJsonValue(curJson);
            node.ReadValue(valueReader);
        }        
    }

private:
    struct NodeRef
    {
        Node* node;
        json::const_pointer jsonPtr;

        NodeRef(Node* node, json::const_pointer jsonPtr) :
            node(node), jsonPtr(jsonPtr)
        {
        }
    };

    std::stack<NodeRef> stack;
    json::const_pointer curJson;
    JsonValueReader valueReader;
};

/**
 * Similar to JsonDeserializerImpl, but instead of writing the json data to the node,
 * it checks that the json data satisfies the node's validation criteria.
 */
class JsonValidatorImpl : public INodeVisitorConst
{
public:
    /**
     * Traverses through the json, checking validation criteria for each value.
     * Currently throws if a discrepancy is identified.
     * Future improvements include container validation support and a returned list of errors (as opposed to throwing at the first one).
     */
    void Validate(const Node& node, const nlohmann::json& json)
    {
        stack.emplace(&node, &json);

        while (!stack.empty())
        {
            const Node& curNode = *stack.top().node;
            curJson = stack.top().jsonPtr;
            stack.pop();

            curNode.Accept(*this);
        }
    }

    void Visit(const Structure& node) override
    {
        std::vector<std::string> allNames;
        for (auto it = node.begin(); it != node.end(); ++it)
        {
            auto& name = it->first;
            auto& node = it->second;

            allNames.push_back(name);

            if (node.WriteOnly())
            {
                continue;
            }

            json::const_pointer jsonItem = nullptr;

            if (curJson)
            {
                auto jsonIt = curJson->find(name);
                if (jsonIt != curJson->end())
                {
                    jsonItem = &jsonIt.value();
                }
            }

            stack.emplace(&node, jsonItem);
        }

        // TODO: add structure validation once implemented
        // useful args: allNames
        // node.CheckValidation(args...)
    }

    void Visit(const Internal::ArrayBase& node) override
    {
        if (curJson)
        {
            // Modified based on GOS-1337
            GoThrowIf(!curJson->is_array(), kERROR_PARAMETER);

            // GOS-3739 - Depending if the json array has less or more entries
            // than the node, we need to index properly.
            // Only validate to the smallest size array to avoid out of bounds indexing.
            size_t arrayCount = node.Count() > curJson->size() ? curJson->size() : node.Count();
            for (size_t i = 0; i < arrayCount; ++i)
            {
                stack.emplace(node.At(i), &curJson->at(i));
            }
        }
        else
        {
            for (size_t i = 0; i < node.Count(); ++i)
            {
                stack.emplace(node.At(i), nullptr);
            }
        }

        // TODO: add array validation once implemented
        // useful args: curJson->size()
        // node.CheckValidation(args...)
    }

    void Visit(const Internal::ReferenceArray& node) override
    {
        // ReferenceArray is used only in schemas, which are never read.
    }

    void Visit(const Internal::ValueBase& node) override
    {
        if (curJson)
        {
            valueReader.SetJsonValue(curJson);
            node.ValidateValue(valueReader);
        }        
    }

private:
    struct NodeRef
    {
        const Node* node;
        json::const_pointer jsonPtr;

        NodeRef(const Node* node, json::const_pointer jsonPtr) :
            node(node), jsonPtr(jsonPtr)
        {
        }
    };

    std::stack<NodeRef> stack;
    json::const_pointer curJson;
    JsonValueReader valueReader;
};

nlohmann::json JsonSerializer::Serialize(const Node& node)
{
    JsonSerializerImpl serializer;

    return serializer.Serialize(node);
}

void JsonSerializer::Deserialize(Node& node, const nlohmann::json& json)
{
    JsonDeserializerImpl deserializer;

    deserializer.Deserialize(node, json);
}

void JsonSerializer::ValidateJson(const Node& node, const nlohmann::json& json)
{
    JsonValidatorImpl validator;

    validator.Validate(node, json);
}

nlohmann::json JsonSerializer::SerializeSchema(const Node& node)
{
    if (node.Schema())
    {
        return Serialize(*node.Schema());
    }

    return nullptr;
}

ByteVector JsonSerializer::DecipherBinaryJson(const nlohmann::json& binary)
{
    ByteVector buffer;
    if (binary.is_binary())
    {
        // Received via msgpack. Parse binary content.
        nlohmann::json::binary_t bytes = binary.get<nlohmann::json::binary_t>();
        buffer = ByteVector(bytes);
    }
    else if (binary.is_object())
    {
        // Received as normal json. Parse bytes sub-element.
        buffer = binary["bytes"].get<ByteVector>();
    }
    else
    {
        GoThrowMsg(kERROR_PARAMETER, "Misformatted binary json object.");
    }
    return buffer;
}

}
};
