/**
 * @file    GoDiscoveryClient.cpp
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
#include <GoPxLSdk/Def.h>
#include <GoPxLSdk/GoDiscoveryClient.h>


namespace GoPxLSdk 
{

void GoDiscoveryClient::BlockingDiscover(k64u timeout)
{
    DiscoveryBroadcastHeader header;
    kUdpClient receiver = kNULL;
    Byte buffer[MAX_MESSAGE_SIZE];
    kSize readSize = 0;
    kTimer timer = kNULL;
    
    instances.clear();
    try
    {
        GoTest(kTimer_Construct(&timer, kAlloc_App()));

        // Fill the message header that will be broadcasted to the server.
        header.length = sizeof(header);
        header.messageId = GOPXL_DISCOVERY_MESSAGE_DISCOVER;
        header.signature = GOPXL_BROADCAST_SIGNATURE;

        ConstructReceiver(&receiver);
        Broadcast(&header, sizeof(header));
        GoTest(kTimer_Start(timer, timeout));
        do
        {
            if (kSuccess(kUdpClient_Receive(receiver, kNULL, &readSize, kTimer_Remaining(timer))))
            {
                GoThrowIf(readSize > MAX_MESSAGE_SIZE, kERROR_MEMORY);
                if (kSuccess(kStream_Read(receiver, buffer, readSize)))
                { 
                    try
                    {
                        ParseReply(buffer);
                    }
                    catch (const std::exception&)
                    {
                        // We want to continue to receive UDP messages even if ParseReply fails
                        // to potentially parse more GoPxL instances.
                        continue;
                    }
                }
            }
        }
        while (!kTimer_IsExpired(timer));  
        GoTest(kTimer_Stop(timer));
    }
    catch (const Go::Exception&)
    {
        kDestroyRef(&receiver);
        kDestroyRef(&timer);
        GoRethrow("Failed in discovering GoPxL instances.");
    }
    kDestroyRef(&receiver);
    kDestroyRef(&timer);
}

const std::vector<GoInstance>& GoDiscoveryClient::InstanceList()
{
    return instances;
}

const GoInstance* GoDiscoveryClient::Instance(std::string appId)
{
    for (std::vector<GoInstance>::size_type  i = 0; i < instances.size(); i++) {
        if (instances.at(i).GetAppId() == appId) 
        {
            return &instances.at(i);
        }
    }
    return NULL;
}

void GoDiscoveryClient::ConstructReceiver(kUdpClient* receiver)
{
    kUdpClient localReceiver = kNULL;

    try
    {
        GoThrowIf(receiver == nullptr, kERROR_PARAMETER);
        GoTest(kUdpClient_Construct(&localReceiver, kIP_VERSION_4, kAlloc_App()));
        GoTest(kUdpClient_EnableBroadcast(localReceiver, kTRUE));
        GoTest(kUdpClient_EnableReuseAddress(localReceiver, kTRUE));
        GoTest(kUdpClient_SetReadBuffers(localReceiver, CLIENT_SOCKET_SIZE, CLIENT_BUFFER_SIZE));
        GoTest(kUdpClient_Bind(localReceiver, kIpAddress_AnyV4(), GOPXL_RESERVED_PORT_DISCOVERY_PROTOCOL));
        
        *receiver = localReceiver;
    }
    catch (const Go::Exception&)
    {
        kObject_Destroy(localReceiver);
        GoRethrow("Failed to construct receiver.");
    }

}

void GoDiscoveryClient::ConstructSender(kUdpClient* sender, kIpAddress address)
{
    kUdpClient localSender = kNULL;

    try
    {
        GoThrowIf(sender == nullptr, kERROR_PARAMETER);
        GoTest(kUdpClient_Construct(&localSender, kIP_VERSION_4, kAlloc_App()));
        GoTest(kUdpClient_EnableBroadcast(localSender, kTRUE));
        GoTest(kUdpClient_EnableReuseAddress(localSender, kTRUE));
        GoTest(kUdpClient_SetWriteBuffers(localSender, CLIENT_SOCKET_SIZE, CLIENT_BUFFER_SIZE));
        GoTest(kUdpClient_Bind(localSender, address, kIP_PORT_ANY));

        *sender = localSender;
    }
    catch (const Go::Exception&)
    {
        kObject_Destroy(localSender);
        GoRethrow("Failed to construct sender.");
    }

}

void GoDiscoveryClient::Broadcast(const void * data, k32u length)
{
    kArrayList entries = kNULL;
    kUdpClient sender = kNULL;
    kIpEntry loopback = { 0 };
    const kChar* LOOPBACK_NAME = "loopback";

    try
    {
        GoTest(kArrayList_Construct(&entries, kTypeOf(kVoid), 0, kAlloc_App()));
        GoTest(kNetwork_LocalIpInterfaces(entries));
        // Include loopback addresses if using virtual devices.
        loopback.address = kIpAddress_LoopbackV4();
        GoTest(kStrCopy(loopback.name, sizeof(loopback.name), LOOPBACK_NAME));
        GoTest(kArrayList_AddT(entries, &loopback));

        for (kSize i = 0; i < kArrayList_Count(entries); i++)
        {
            kIpEntry* entry = kArrayList_AtT(entries, i, kIpEntry);

            // We want to continue loop even if ConstructSender fails to create a sender for that 
            // interface instead of skipping all remaining interfaces.
            try
            {
                ConstructSender(&sender, entry->address);
            }
            catch (const Go::Exception&)
            {
                continue;
            }

            GoTest(kStream_Write(sender, data, length));
            GoTest(kUdpClient_Send(sender, kIpAddress_BroadcastV4(), GOPXL_RESERVED_PORT_DISCOVERY_PROTOCOL, kINFINITE, kTRUE));
            GoTest(kDestroyRef(&sender));
        }
    }
    catch (const Go::Exception&)
    {
        kObject_Destroy(entries);
        kDestroyRef(&sender);
        GoRethrow("Failed to broadcast message.");
    }
    kObject_Destroy(entries);
    kDestroyRef(&sender);
}

void GoDiscoveryClient::ParseReply(Byte bytes[])
{
    GoInstance instanceInfo;
    int pos = sizeof(DiscoveryServerHeader);
    kChar* charArr = (kChar*)bytes+pos;

    // Extract the header of the message that was received from server.
    const auto serverHeader = reinterpret_cast<DiscoveryServerHeader*>(bytes);

    // Ignore any discover messages.
    if (serverHeader->messageId == GOPXL_DISCOVERY_MESSAGE_DISCOVER) 
    {
        return;
    }
    GoThrowIf(serverHeader->messageId != GOPXL_DISCOVERY_MESSAGE_ANNOUNCE || serverHeader->signature != GOPXL_BROADCAST_SIGNATURE, kERROR);

    // Convert the message payload to a string.
    std::string instanceInfoStr(charArr);

    // Convert the string to json object.
    auto instanceInfoJson = GoJson::ParseString(instanceInfoStr);

    // Store the information of the json object into a GoInstance object.
    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_APP_ID))
    {
        // If this instance is a repeat, do not store information.
        if (Instance(instanceInfoJson.Get<std::string>(GOPXL_DISCOVERY_APP_ID)) != NULL) {return;}
        instanceInfo.SetAppId(instanceInfoJson.Get<std::string>(GOPXL_DISCOVERY_APP_ID));
    }
    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_SERIAL_NUMBER) && 
        !instanceInfoJson.Get<std::string>(GOPXL_DISCOVERY_SERIAL_NUMBER).empty())
    {
        k32u serialNumber;
        GoTest(k32u_Parse(&serialNumber, instanceInfoJson.Get<std::string>(GOPXL_DISCOVERY_SERIAL_NUMBER).c_str()));
        instanceInfo.SetSerialNumber(serialNumber);
    }
    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_ADDRESS) && 
        !instanceInfoJson.Get<std::string>(GOPXL_DISCOVERY_ADDRESS).empty())
    {
        kIpAddress ipAddress;
        GoTest(kIpAddress_Parse(&ipAddress, instanceInfoJson.Get<std::string>(GOPXL_DISCOVERY_ADDRESS).c_str()));
        instanceInfo.SetIpAddress(ipAddress);
    }
    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_GATEWAY) && 
        !instanceInfoJson.Get<std::string>(GOPXL_DISCOVERY_GATEWAY).empty()) 
    {
        kIpAddress gateway;
        GoTest(kIpAddress_Parse(&gateway, instanceInfoJson.Get<std::string>(GOPXL_DISCOVERY_GATEWAY).c_str()));
        instanceInfo.SetGateway(gateway);
    }
    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_MASK) && 
        !instanceInfoJson.Get<std::string>(GOPXL_DISCOVERY_MASK).empty())
    {
        kIpAddress mask;
        GoTest(kIpAddress_Parse(&mask, instanceInfoJson.Get<std::string>(GOPXL_DISCOVERY_MASK).c_str()));
        instanceInfo.SetMask(mask);
    }

    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_CONTROL_PORT)) 
    {
        instanceInfo.SetControlPort(instanceInfoJson.Get<k32u>(GOPXL_DISCOVERY_CONTROL_PORT));
    }
    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_GDP_PORT)) 
    {
        instanceInfo.SetGdpPort(instanceInfoJson.Get<k32u>(GOPXL_DISCOVERY_GDP_PORT));
    }
    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_WEB_PORT)) 
    {
        instanceInfo.SetWebPort(instanceInfoJson.Get<k32u>(GOPXL_DISCOVERY_WEB_PORT));
    }
    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_APP_NAME))
    {
        instanceInfo.SetAppName(instanceInfoJson.Get<std::string>(GOPXL_DISCOVERY_APP_NAME));
    }
    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_APP_VERSION))
    {
        instanceInfo.SetAppVersion(instanceInfoJson.Get<std::string>(GOPXL_DISCOVERY_APP_VERSION));
    }
    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_DEVICE_MODEL))
    {
        instanceInfo.SetDeviceModel(instanceInfoJson.Get<std::string>(GOPXL_DISCOVERY_DEVICE_MODEL));
    }
    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_DHCP))
    {
        instanceInfo.SetIsDhcp(instanceInfoJson.Get<bool>(GOPXL_DISCOVERY_DHCP));
    }
    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_IS_REMOTE))
    {
        instanceInfo.SetIsRemote(instanceInfoJson.Get<bool>(GOPXL_DISCOVERY_IS_REMOTE));
    }
    if (instanceInfoJson.HasKey(GOPXL_DISCOVERY_ADDRESS_CONFLICT))
    {
        instanceInfo.SetIsAddressConflict(instanceInfoJson.Get<bool>(GOPXL_DISCOVERY_ADDRESS_CONFLICT));
    }

    instances.push_back(instanceInfo);
}

}