/**
 * @file    GoGdpClient.h
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
#ifndef GO_PXL_SDK_GOGDPCLIENT_H
#define GO_PXL_SDK_GOGDPCLIENT_H

#include <kApi/Io/kNetwork.h>
#include <kApi/Io/kTcpClient.h>

#include <kApi/Threads/kThread.h>
#include <kApi/Threads/kLock.h>
#include <kApi/Threads/kTimer.h>

#include <GoApi/GoApi.h>
#include <GoApi/Threads/MsgQueue.h>
#include <GoApi/Threads/Locker.h>

#include <GoPxLSdk/Def.h>
#include <GoPxLSdk/GoDataSet.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpMsgDef.h>

namespace GoPxLSdk
{
    class GoPxLSdkClass GoGdpClient
    {
    public:
        /**
        * Constructs GoGdpClient.
        *
        * @public                @memberof GoGdpClient
        * @version               Introduced in 0.2.1.53
        * @throws Go::Exception if unable to create client.
        */
        GoGdpClient();
        ~GoGdpClient();

        const int CONNECT_TIMEOUT = 5000000;
        const int CANCEL_QUERY_INTERVAL = 100000;

        /**
        * Connect to GdpServer.
        *
        * @public                @memberof GoGdpClient
        * @version               Introduced in 0.2.1.53
        * @param ipAddr ip addresss.
        * @param port default DEFAULT_GDP_PORT.
        * @throws Go::Exception if unable to connect.
        */
        void Connect(kIpAddress ipAddr, k32u port = GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT);

        /**
        * Close the tcp connection.
        *
        * @public                @memberof GoGdpClient
        * @version               Introduced in 0.2.1.53
        * @throws Go::Exception if unable to close connection.
        */
        void Close();

        /**
        * Checks if the client is connected or not.
        *
        * @public                @memberof GoGdpClient
        * @version               Introduced in 0.2.1.53
        * @return True if connected, false otherwise.
        */
        bool IsConnected();

        /**
        * Sync Receive data from gdp server.
        *
        * @public                @memberof GoGdpClient
        * @version               Introduced in 0.2.1.53
        * @param receiveTimeout Set time out.
        * @throws Go::Exception if unsuccessful.
        */
        void ReceiveDataSync(k64u receiveTimeout);

        /**
        * Async Receive data from gdp server.
        *
        * @public                @memberof GoGdpClient
        * @version               Introduced in 0.2.1.53
        * @param func   Callback function to get data.
        * @throws Go::Exception if unsuccessful.
        */
        void ReceiveDataAsync(std::function<void(const GoDataSet& receivedDataSet)>& func);

        /**
        * Get the data set received from the sensor over the Gocator Data Protocol.
        *
        * @public                @memberof GoGdpClient
        * @version               Introduced in 0.2.1.53
        * @return the pointer of dataSet.
        */
        const GoDataSet& DataSet() const;

    private:
        /**
        * Get the gdp msg type.
        *
        * @private               @memberof GoGdpClient
        * @version               Introduced in 0.2.1.53
        * @param type Pass a reference to get the msg type.
        * @return Go::Exception if unsuccessful.
        */
        void GetGdpMsgType(MessageType& type);
        void InitTcpClient();
        void InitThreadAndTimer();
        std::function<void(const GoDataSet& pDataSet)> func;

        bool isConnected = false;
        bool async = false;
        bool isCompleteSet;

        Go::Object<kTcpClient> socket;
        Go::Object<kSerializer> serializer;
        Go::Object<kTimer> timer;
        Go::Object<kThread> receiveThread;
        static kStatus kCall OnDataReceive(void* context);
        GoDataSet dataSet;

        static kStatus kCall DataThreadEntry(void* context);
        Go::MsgQueue<GoDataSet> dataQueue;
        Go::Object<kThread> dataThread;

    };
}


#endif


