/**
 * @file    GoGdpImage.cpp
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
#include <GoPxLSdk/GoGdpMsg/GoGdpImage.h>
#include <kApi/Data/kImage.h>

namespace GoPxLSdk
{
    GoGdpImage::GoGdpImage()
        : GoGdpMsg(MessageType::IMAGE)
    {
        image.Reset();
    }

    void GoGdpImage::Deserialize(kSerializer serializer)
    {
        kSize rowSize;
        bool needEndRead = false;

        GoGdpMsg::Deserialize(serializer);

        try
        {
            k16u reserved16;
            k8u byteValue;

            GoTest(kSerializer_BeginRead(serializer, kTypeOf(k16u), kTRUE));
            needEndRead = true;
            GoTest(kSerializer_Read32u(serializer, &height));
            GoTest(kSerializer_Read32u(serializer, &width));
            GoTest(kSerializer_Read32u(serializer, &pixelSize));
            GoTest(kSerializer_Read32s(serializer, &pixelFormat));
            GoTest(kSerializer_Read32s(serializer, &colorFilter));
            GoTest(kSerializer_Read16u(serializer, &reserved16));

            GoTest(kSerializer_Read32f(serializer, &exposure));
            GoTest(kSerializer_Read8u(serializer, &byteValue));
            flippedX = (byteValue > 0);
            GoTest(kSerializer_Read8u(serializer, &byteValue));
            flippedY = (byteValue > 0);
            GoTest(kSerializer_Read8u(serializer, &byteValue));
            columnBased = (byteValue > 0);

            GoTest(kSerializer_EndRead(serializer));
            needEndRead = false;

            if ((pixelFormat == kPIXEL_FORMAT_8BPP_GREYSCALE) || (pixelFormat == kPIXEL_FORMAT_8BPP_CFA))
            {
                GoTest(kImage_Construct(image.Ref(), kTypeOf(k8u), width, height, kAlloc_App()));

                rowSize = pixelSize * width;

                for (k32u i = 0; i < height; ++i)
                {
                    void* pixels = kImage_RowAtT(image, i, k8u);
                    GoTest(kSerializer_ReadByteArray(serializer, pixels, rowSize));
                }
            }
            else if (pixelFormat == kPIXEL_FORMAT_8BPC_BGRX)
            {
                GoTest(kImage_Construct(image.Ref(), kTypeOf(kRgb), width, height, kAlloc_App()));

                for (k32u i = 0; i < height; ++i)
                {
                    void* pixels = kImage_RowAtT(image, i, kRgb);
                    GoTest(kSerializer_Read8uArray(serializer, (k8u*)pixels, 4 * width));
                }
            }
        }
        catch (const Go::Exception&)
        {
            if (needEndRead)
            {
                kSerializer_EndRead(serializer);
            }

            GoRethrow("Failed to deserialize Image GDP message.");
        }
    }

    const k32u GoGdpImage::Height() const
    {
        return height;
    }

    const k32u GoGdpImage::Width() const
    {
        return width;
    }

    const k32u GoGdpImage::PixelSize() const
    {
        return pixelSize;
    }

    const kPixelFormat GoGdpImage::PixelFormat() const
    {
        return pixelFormat;
    }

    const kCfa GoGdpImage::ColorFilter() const
    {
        return colorFilter;
    }

    const k32f GoGdpImage::Exposure() const
    {
        return exposure;
    }

    const bool GoGdpImage::FlippedX() const
    {
        return flippedX;
    }

    const bool GoGdpImage::FlippedY() const
    {
        return flippedY;
    }

    const bool GoGdpImage::ColumnBased() const
    {
        return columnBased;
    }

    const kImage GoGdpImage::Image() const
    {
        return image.Get();
    }

}
