/*
 * @file            AlignSensor.cpp
 * Objective        Connect to a Gocator sensor system to perform alignment.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

#include <fstream>

#include <kApi/kApiDef.h>
#include <kApi/Threads/kThread.h>
#include <GoApi/GoApiLib.h>
#include <GoPxLSdk/GoResponse.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoChannelError.h>
#include <GoPxLSdk/GoRequestError.h>

using std::string;
using namespace GoPxLSdk;

#define SENSOR_IP "192.168.1.10"

const string ENGINE_ID = "g2xxx";
const string ALIGN_COMMAND_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/commands/align";
const string UNALIGN_COMMAND_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/commands/unalign";
const string ALIGNMENT_STATE_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/alignment";
const int ERROR_STATUS = -1;
const int OK_STATUS = 0;
const int REST_COMMAND_TIMEOUT = 3'000'000; // 3 seconds

int AlignSensor()
{
    GoSystem system;
    kStatus status;
    kAssembly goApiLib = kNULL;
    GoJson response;
    GoJson payload;

    // Constructs GoPxL API core framework.
    if ((status = GoApiLib_Construct(&goApiLib)) != kOK)
    {
        std::cout << "Error: " << status << " - Failed to construct GoApiLib." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to a sensor system.
    kIpAddress systemIpAddress;
    kIpAddress_Parse(&systemIpAddress, SENSOR_IP);
    system = GoSystem(systemIpAddress, GO_PXL_SDK_DEFAULT_CONTROL_PORT);

    kChar ipAddress[16];
    kIpAddress_Format(system.Address(), ipAddress, sizeof(ipAddress));

    try
    {
        system.Connect();
        std::cout << "\nConnected to " << ipAddress << ":" << system.ControlPort() << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Connection failed. Check if sensor is powered on and connected "
                                 "also check if you are using the correct IP address and port number."
                  << std::endl;
        return ERROR_STATUS;
    }

    // Ensure job is saved before performing alignment.
    payload = GoJson("{\"name\" : \"SDK_alignment_sample\" }");
    system.Client().Call("/jobs/commands/save", payload).CheckResponse(REST_COMMAND_TIMEOUT);
    std::cout << "\nSaved job as " << payload.At("name") << std::endl;

    // Clear alignment.
    std::cout << "\nClearing alignment..." << std::endl;
    system.Client().Call(UNALIGN_COMMAND_PATH).CheckResponse(REST_COMMAND_TIMEOUT);
    std::cout << "Alignment cleared." << std::endl;

    // Align the sensor.
    const std::map<int, string> alignmentStates{{0, "Not Aligned"}, {1, "Aligned"}, {2, "Aligning"}};
    std::cout << "\nExecuting alignment..." << std::endl;
    try
    {
        system.Client().Call(ALIGN_COMMAND_PATH).CheckResponse(REST_COMMAND_TIMEOUT);

        // Wait while alignment is in progress.
        bool aligning = true;
        int alignmentState = 0;
        while (aligning)
        {
            kThread_Sleep(1'000'000);
            response = system.Client().Read(ALIGNMENT_STATE_PATH).GetResponse(REST_COMMAND_TIMEOUT).Payload();
            alignmentState = response.At("/alignState").Get<int>();
            if (alignmentState != 2)
            {
                aligning = false;
            }
        }
        std::cout << "Alignment state: " << alignmentStates.at(alignmentState) << std::endl;
        std::cout << "Alignment configuration: " << response.At("/alignConfig") << std::endl;

        if (alignmentState == 0)
        {
            std::cout << "\nAlignment failed. Check alignment configuration on sensor." << std::endl;
        }
    }
    catch (const std::exception& e)
    {
        std::cout << "\nAlignment failed. Check API path, sensor state and availability of scan data." << std::endl;
    }

    system.Disconnect();
    return OK_STATUS;
}

int main(int argc, char **argv)
{
    int status;
    try
    {
        status = AlignSensor();
    }
    catch (const GoRequestError &e)
    {
        GoJson errors = e.GetResponse().Payload().At("errors");

        string errorMessage = "";

        if (!errors.Empty())
        {
            errorMessage.append(errors.At("/0/status").ToString());
            errorMessage.append(" ");
            errorMessage.append(errors.At("/0/description").ToString());
        }

        std::cerr << "GoRequestError: " << errorMessage << std::endl;
        std::cerr << "Error sending a REST command to " << e.GetResponse().Path() << std::endl;
        status = ERROR_STATUS;
    }
    catch (const GoChannelError &e)
    {
        std::cerr << "Error: " << e.what() << " - Check sensor status and make sure it is connected." << std::endl;
        status = ERROR_STATUS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        status = ERROR_STATUS;
    }

    std::cout << "\nPress ENTER to exit the program..." << std::endl;
    getchar();

    return status;
}