/*
 * @file            ConfigureTool.cpp
 * Objective        Connect to a Gocator sensor system, add a tool, change configurations of the tool, and enable measurement data output.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

#include <kApi/kApiDef.h>
#include <GoApi/GoApiLib.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoChannelError.h>
#include <GoPxLSdk/GoRequestError.h>

using std::string;
using namespace GoPxLSdk;

#define SENSOR_IP       "192.168.1.10"

const string TOOLS_PATH = "/tools/";
const string DATA_SOURCE_PATH = "/dataSources/";
const string GOCATOR_CONTROL_PATH = "/controls/gocator";
const string GOCATOR_OUTPUT_PATH = "/controls/gocator/outputs/commands/add";
const int ERROR_STATUS = -1;
const int OK_STATUS = 0;
const int REST_COMMAND_TIMEOUT = 3'000'000; // 3 seconds.

int ConfigureTool()
{
    GoSystem system;
    kStatus status;
    kAssembly goApiLib = kNULL;
    GoJson response;

    // Constructs GoPxL API core framework.
    if ((status = GoApiLib_Construct(&goApiLib)) != kOK)
    {
        std::cout << "Error: " << status << " - Failed to construct GoApiLib." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to a sensor system.
    kIpAddress systemIpAddress;
    kIpAddress_Parse(&systemIpAddress, SENSOR_IP);
    system = GoSystem(systemIpAddress, GO_PXL_SDK_DEFAULT_CONTROL_PORT);

    kChar ipAddress[16];
    kIpAddress_Format(system.Address(), ipAddress, sizeof(ipAddress));

    try
    {
        system.Connect();
        std::cout << "\nConnected to " << ipAddress << ":" << system.ControlPort() << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Connection failed. Check if sensor is powered on and connected "
            "also check if you are using the correct IP address and port number." << std::endl;
        return ERROR_STATUS;
    }

    // Configuration 1: Add a "Profile Bounding Box" tool.
    auto payload = GoJson("{\"type\" : \"ProfileBoundingBox\", \"autoConnect\": false, \"position\": 0}");
    system.Client().Create(TOOLS_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
    std::cout << "\nNew tool added: " << payload.At("type") << std::endl;

    // Configuration 2: Update the tool id.
    payload = GoJson("{\"extId\": \"ProfileBoundingBox-demo\"}");
    try
    {
        system.Client().Update("tools/ProfileBoundingBox-0", payload).CheckResponse(REST_COMMAND_TIMEOUT);
        std::cout << "\nTool id updated: " << payload.At("extId") << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to update tool id. Check the existing tool id and try again." << std::endl;
        return ERROR_STATUS;
    }

    // Configuration 3: Connect tool to a data source.
    // Read all available outputs (data sources).
    response = system.Client().Read(DATA_SOURCE_PATH).GetResponse(REST_COMMAND_TIMEOUT).Payload();
    auto dataSources = response.At("_embedded/item");

    string dataSourceId;
    std::vector<string> dataSourceIds;
    std::cout << "\nData sources:" << std::endl;
    for (auto it = dataSources.Begin(); it != dataSources.End(); it.operator++(0))
    {
        dataSourceId = it.Value().At("/_links/self/href").Get<string>();
        std::cout << dataSourceId << std::endl;
        dataSourceIds.push_back(dataSourceId);
    }
    // Read all available tool inputs.
    response = system.Client().Read("/tools/extId=ProfileBoundingBox-demo/inputs").GetResponse().Payload();
    auto inputs = response.At("_embedded/item");

    string inputId;
    std::vector<string> inputIds;
    std::cout << "\nTool inputs:" << std::endl;
    for (auto it = inputs.Begin(); it != inputs.End(); it.operator++(0))
    {
        inputId = it.Value().At("/_links/self/href").Get<string>();
        std::cout << inputId << std::endl;
        inputIds.push_back(inputId);
    }

    // Connect tool input to a data source.
    string profileInputPath = "/tools/extId=ProfileBoundingBox-demo/inputs/ProfileInput";
    payload = GoJson("{\"dataSource\": \"scan:g2xxx:scanner-0:topUniformProfile\"}"); // TODO: look at surface sample
    try
    {
        system.Client().Update(profileInputPath, payload).CheckResponse(REST_COMMAND_TIMEOUT);
        std::cout << "\nTool input connected." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to update tool input. Check the tool id and API path." << std::endl;
        return ERROR_STATUS;
    }

    // Configuration 4: Enable a measurement output.
    payload = GoJson("{\"enabled\": true}");
    string output = "X";
    string MEASUREMENT_OUTPUT_PATH = "/tools/extId=ProfileBoundingBox-demo/outputs/" + output;
    try
    {
        system.Client().Update(MEASUREMENT_OUTPUT_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
        std::cout << "\nMeasurement output enabled." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to enable the output. Check the tool id and API path." << std::endl;
        return ERROR_STATUS;
    }
    // Configuration 5: Send measurement value over GDP.
    // Read measurement output.
    string METRICS_PATH = "/tools/extId=ProfileBoundingBox-demo/metrics";
    
    try
    {
        response = system.Client().Read(METRICS_PATH).GetResponse().Payload();
        std::cout << "\nMeasurement value: " << response.At("/outputsByExtId/" + output + "/value") << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to read value. Check the API path and ensure scan data is available." << std::endl;
        return ERROR_STATUS;
    }
    

    // Enable Gocator Protocol.
    payload = GoJson("{\"enabled\": true}");
    system.Client().Update(GOCATOR_CONTROL_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
    std::cout << "\nGocator Protocol enabled." << std::endl;

    // Send the value.
    payload = GoJson("{\"source\": \"tools:ProfileBoundingBox-0:outputs:X\", \"outputId\": 1, \"autoShift\": true}");
    try
    {
        system.Client().Call(GOCATOR_OUTPUT_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
        std::cout << "\nMeasurement value added to Gocator output." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to send the data. Check the tool name and API path." << std::endl;
        return ERROR_STATUS;
    }

    system.Disconnect();

    return OK_STATUS;
}

int main(int argc, char** argv)
{
    int status;

    try
    {
        status = ConfigureTool();
    }
    catch (const GoRequestError& e)
    {
        string errorMessage = "";

        const GoJson& payload = e.GetResponse().Payload();
        if (!payload.Empty() && payload.HasKey("errors"))
        {
            GoJson errors = payload.At("errors");
            errorMessage.append(errors.At("/0/status").ToString());
            errorMessage.append(" ");
            errorMessage.append(errors.At("/0/description").ToString());
        }

        std::cerr << "GoRequestError: " << errorMessage << std::endl;
        std::cerr << "Error sending a REST command to " << e.GetResponse().Path() << std::endl;
        status = ERROR_STATUS;
    }
    catch (const GoChannelError& e)
    {
        std::cerr << "Error: " << e.what() << " - Check sensor status and make sure it is connected." << std::endl;
        status = ERROR_STATUS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        status = ERROR_STATUS;
    }

    std::cout << "\nPress ENTER to exit the program..." << std::endl;
    getchar();

    return status;
}
