/*
 * @file            MultiSensorLayout.cpp
 * Objective        Connect to the main sensor, add a sensor to sensor group, and change the layout of the group.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

#include <kApi/kApiDef.h>
#include <GoApi/GoApiLib.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoRequestError.h>
#include <GoPxLSdk/GoChannelError.h>
#include <GoPxLSdk/GoInstance.h>

using std::string;
using namespace GoPxLSdk;

#define SENSOR_1_IP "192.168.1.10"
#define SENSOR_2_IP "192.168.1.11"

const string SENSOR_1_SERIAL_NO = "62048";
const string SENSOR_2_SERIAL_NO = "62984";
const string ENGINE_ID = "g2xxx";
const string ENGINE_PATH = "/scan/engines/" + ENGINE_ID;
const string SCANNERS_PATH = "/scan/engines/" + ENGINE_ID + "/scanners";
const string SENSORS_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/sensors";
const string VISIBLE_SENSORS_PATH = "/scan/engines/" + ENGINE_ID + "/visibleSensors";
const string SENSOR_LAYOUT_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/layout";
const int ERROR_STATUS = -1;
const int OK_STATUS = 0;
const int REST_COMMAND_TIMEOUT = 3'000'000;
const int RECEIVE_DATA_TIMEOUT = 3'000'000;

string parseGoRequestError(const GoRequestError &e)
{
    GoJson errors = e.GetResponse().Payload().At("errors");

    string errorMessage = "";

    if (!errors.Empty())
    {
        errorMessage.append("Error: ");
        errorMessage.append(errors.At("/0/status").ToString());
        errorMessage.append(" ");
        errorMessage.append(errors.At("/0/description").ToString());
    }

    return errorMessage;
}

void printSensorLayoutInfo(GoJson response)
{
    GoJson sensors = response.At("/grid/sensors");
    std::cout << "\n"
              << string(32, '*') << std::endl;
    for (auto sensor = sensors.Begin(); sensor < sensors.End(); sensor++)
    {
        const GoJson &sensorLayoutInfo = sensor.Value();
        std::cout << "Sensor ID: " << sensorLayoutInfo.At("sensorId") << std::endl;
        std::cout << "Row: " << sensorLayoutInfo.At("row") << ", Column: " << sensorLayoutInfo.At("column") << std::endl;
        std::cout << "Orientation: " << sensorLayoutInfo.At("orientation") << std::endl;
        std::cout << "Multiplexing Bank: " << sensorLayoutInfo.At("multiplexingBank") << "\n"
                  << std::endl;
    }
}

int MultiSensorLayout()
{
    GoSystem system;
    kStatus status;
    kAssembly goApiLib = kNULL;
    GoJson response;
    GoJson payload;

    // Constructs GoPxL API core framework.
    if ((status = GoApiLib_Construct(&goApiLib)) != kOK)
    {
        std::cout << "Error: " << status << " - Failed to construct GoApiLib." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to a sensor system.
    kIpAddress systemIpAddress;
    kIpAddress_Parse(&systemIpAddress, SENSOR_1_IP);
    system = GoSystem(systemIpAddress, GO_PXL_SDK_DEFAULT_CONTROL_PORT);
    kChar ipAddress[16];
    kIpAddress_Format(system.Address(), ipAddress, sizeof(ipAddress));
    try
    {
        system.Connect();
        std::cout << "\nConnected to " << ipAddress << ":" << system.ControlPort() << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Connection failed. Check if sensor is powered on and connected "
                                 "also check if you are using the correct IP address and port number."
                  << std::endl;
        return ERROR_STATUS;
    }

    // Check if a sensor group exists.
    response = system.Client().Read(ENGINE_PATH).GetResponse(REST_COMMAND_TIMEOUT).Payload();
    if (response.At("/_embedded/go:scanner").Empty())
    {
        std::cout << "\nScanner not present, creating scanner" << std::endl;
        system.Client().Create(SCANNERS_PATH);
    }
    else
    {
        std::cout << "\nA Sensor Group exists." << std::endl;
    }

    // Add a sensor to the sensor group.
    std::vector<string> sensorSerials = {SENSOR_2_SERIAL_NO};
    for each (string serial in sensorSerials)
    {
        payload = GoJson("{\"serialNumber\":\"" + serial + "\"}");
        try
        {
            system.Client().Create(SENSORS_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
            std::cout << "\nAdded " << serial << " to the sensor group." << std::endl;
        }
        catch (const std::exception& e)
        {
            std::cerr << e.what() << " - Failed to add sensor " << serial << ". Check the API path & serial number and make sure firmware is compatible and that sensor is not already added." << std::endl;
            return ERROR_STATUS;
        }
    }

    // Output the sensor group.
    response = system.Client().Read(SENSOR_LAYOUT_PATH).GetResponse(REST_COMMAND_TIMEOUT).Payload();
    printSensorLayoutInfo(response);

    // Change the sensor group layout.
    payload = response;
    payload.Replace("/grid/sensors/2/row", 0);
    payload.Replace("/grid/sensors/2/column", 1);
    payload.Replace("/grid/sensors/2/orientation", 0);
    try
    {
        system.Client().Update(SENSOR_LAYOUT_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
        std::cout << "Updated sensor layout" << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to update the layout." << std::endl;
        return ERROR_STATUS;
    }

    // Output the current layout of the sensors in the sensor group.
    response = system.Client().Read(SENSOR_LAYOUT_PATH).GetResponse(REST_COMMAND_TIMEOUT).Payload();
    printSensorLayoutInfo(response);

    // Remove the senors from group.
    const string SENSOR_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/sensors/sensor-0";
    try
    {
        system.Client().Delete(SENSOR_PATH).CheckResponse(REST_COMMAND_TIMEOUT);
        std::cout << "Removed " << SENSOR_PATH << " from sensor group." << std::endl;
    }
    catch (const GoRequestError &e)
    {
        std::cerr << e.what() << "Failed to remove the sensor. Check the API path or the sensor ID." << std::endl;
        return ERROR_STATUS;
    }

    system.Disconnect();
    return OK_STATUS;
}

int main(int argc, char **argv)
{
    int status;

    try
    {
        status = MultiSensorLayout();
    }
    catch (const GoRequestError &e)
    {
        std::cerr << parseGoRequestError(e) << std::endl;
        std::cerr << "GoRequestError - Error sending a REST command to " << e.GetResponse().Path() << std::endl;
        status = ERROR_STATUS;
    }
    catch (const GoChannelError &e)
    {
        std::cerr << "Error: " << e.what() << " - Check sensor status and make sure it is connected." << std::endl;
        status = ERROR_STATUS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        status = ERROR_STATUS;
    }

    std::cout << "\nPress ENTER to exit the program..." << std::endl;
    getchar();

    return status;
}
