/*
 * @file            ReceiveMeasurement.cpp
 * Objective        Connect to the GDP server and receive measurement data.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 *
 * Note: Tool measurement output must be enabled and available. Refer
 * to ConfigureTool.cpp for enabling the measurement outputs.
 */

#include <kApi/kApiDef.h>
#include <GoApi/GoApiLib.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoChannelError.h>
#include <GoPxLSdk/GoRequestError.h>
#include <GoPxLSdk/GoGdpClient.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpStamp.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpMeasurement.h>

using std::string;
using namespace GoPxLSdk;

#define SENSOR_IP "192.168.1.10"

const std::string GOCATOR_CONTROL_PATH = "/controls/gocator";
const int ERROR_STATUS = -1;
const int OK_STATUS = 0;
const int REST_COMMAND_TIMEOUT = 3'000'000; // 3 seconds.
const int RECEIVE_DATA_TIMEOUT = 3'000'000; 

int ReceiveMeasurement()
{
    GoSystem system;
    kStatus status;
    kAssembly goApiLib = kNULL;
    GoJson response;
    GoJson payload;

    // Constructs GoPxL API core framework.
    if ((status = GoApiLib_Construct(&goApiLib)) != kOK)
    {
        std::cout << "Error: " << status << " - Failed to construct GoApiLib." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to a sensor system.
    kIpAddress systemIpAddress;
    kIpAddress_Parse(&systemIpAddress, SENSOR_IP);
    system = GoSystem(systemIpAddress, GO_PXL_SDK_DEFAULT_CONTROL_PORT);

    kChar ipAddress[16];
    kIpAddress_Format(system.Address(), ipAddress, sizeof(ipAddress));

    try
    {
        system.Connect();
        std::cout << "\nConnected to " << ipAddress << ":" << system.ControlPort() << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Connection failed. Check if sensor is powered on and connected "
            "also check if you are using the correct IP address and port number."
            << std::endl;
        return ERROR_STATUS;
    }

    // Run the sensor.
    if (system.RunningState() == GoSystem::State::Ready)
    {
        system.Start();
        std::cout << "\nSystem started." << std::endl;
    }

    // Enable Gocator Protocol.
    payload = GoJson("{\"enabled\": true}");
    try
    {
        system.Client().Update(GOCATOR_CONTROL_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
        std::cout << "\nGocator Protocol enabled." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " Check the timeout value or the API path." << std::endl;
        return ERROR_STATUS;
    }
   
    // Connect to the Gocator Data Protocol (GDP) server.
    auto gdpClient = std::make_unique<GoGdpClient>();
    try
    {
        gdpClient->Connect(system.Address(), GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT);
        std::cout << "\nGocator Protocol connected." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << " - Failed to connect to the GDP server." << std::endl;
        return ERROR_STATUS;
    }

    // Establish a synchronous data transfer.
    try
    {
        gdpClient->ReceiveDataSync(RECEIVE_DATA_TIMEOUT);
        std::cout << "\nSynchronous data transfer established." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to receive data via synchronous connection." << std::endl;
        return ERROR_STATUS;
    }

    // Receive data.
    const GoDataSet& receivedDataSet = gdpClient->DataSet();
    if (receivedDataSet.Count() < 1)
    {
        std::cout << "No data is received." << std::endl; 
        return ERROR_STATUS;
    }
    std::cout << "\nTotal number of messages: " << receivedDataSet.Count() << std::endl;

    // Loop through messages in the dataSet.
    for (size_t i = 0; i < receivedDataSet.Count(); i++)
    {
        std::cout << "\nMessage " << i + 1 << " of " << receivedDataSet.Count() << std::endl;
        std::cout << "Data source: " << receivedDataSet.GdpMsgAt(i).DataSourceId() << std::endl;

        switch (receivedDataSet.GdpMsgAt(i).Type())
        {

        // Get measurement output.
        case GoPxLSdk::MessageType::MEASUREMENT:
        {
            const GoGdpMeasurement& MeasurementMsg = (const GoGdpMeasurement&)receivedDataSet.GdpMsgAt(i);
            std::cout << "Array count: " << MeasurementMsg.ArrayedCount() << std::endl;
            std::cout << "Array index: " << MeasurementMsg.ArrayedIndex() << std::endl;
            std::cout << "Value: " << MeasurementMsg.Value() << std::endl;
            std::cout << "Decision: " << (int)MeasurementMsg.Decision() << std::endl;
            std::cout << "Output ID: " << MeasurementMsg.GdpId() << std::endl;
            break;
        }
        }
    }

    std::cout << "\nPress ENTER to exit the program..." << std::endl;
    getchar();
    gdpClient->Close();
    system.Stop();
    std::cout << "System stopped." << std::endl;
    system.Disconnect();

    return OK_STATUS;

}

int main(int argc, char** argv)
{
    int status;

    try
    {
        status = ReceiveMeasurement();
    }
    catch (const GoRequestError& e)
    {
        std::string errorMessage = "";

        const GoJson& payload = e.GetResponse().Payload();
        if (!payload.Empty() && payload.HasKey("errors"))
        {
            GoJson errors = payload.At("errors");
            errorMessage.append(errors.At("/0/status").ToString());
            errorMessage.append(" ");
            errorMessage.append(errors.At("/0/description").ToString());
        }

        std::cerr << "GoRequestError: " << errorMessage << std::endl;
        std::cerr << "Error sending a REST command to " << e.GetResponse().Path() << std::endl;
        status = ERROR_STATUS;
    }
    catch (const GoChannelError& e)
    {
        std::cerr << "Error: " << e.what() << " - Check sensor status and make sure it is connected." << std::endl;
        status = ERROR_STATUS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        status = ERROR_STATUS;
    }
    return status;
}
