/*
 * @file            ReplayData.cpp
 * Objective        Connect to a Gocator sensor system, record profile scans and download/upload a recording file (.gprec).
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

#include <fstream>
#include <filesystem>

#include <GoApi/GoApiLib.h>
#include <kApi/kApiDef.h>
#include <kApi/Data/kArray1.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoChannelError.h>
#include <GoPxLSdk/GoRequestError.h>
#include <GoPxLSdk/GoGdpClient.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpStamp.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpProfileUniform.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpProfilePointCloud.h>

using std::string;
using namespace GoPxLSdk;

#define SENSOR_IP "192.168.1.10"

const string ENGINE_ID = "g2xxx";
const string SENSOR_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0";
const string START_PATH = "/system/commands/start";
const string STOP_PATH = "/system/commands/stop";
const string RECORDING_PATH = "/replay/recording";
const string BACKUP_PATH = "/system/commands/archive";
const string RESTORE_PATH = "/system/commands/restore";
const string REPLAY_SEEK_PATH = "/replay/commands/seek";
const string REPLAY_PLAYBACK_PATH = "/replay/playback";
const string RECORDING_FILE_PATH = "./sample_recording.gprec";
const int ERROR_STATUS = -1;
const int OK_STATUS = 0;
const int REST_COMMAND_TIMEOUT = 3'000'000; // 3 seconds.
const int RECEIVE_DATA_TIMEOUT = 20'000'000; // Receiving data might require a higher timeout value.

int ReplayData()
{
    GoSystem system;
    kStatus status;
    kAssembly goApiLib = kNULL;
    GoJson response;
    GoJson payload;

    // Constructs GoPxL API core framework.
    if ((status = GoApiLib_Construct(&goApiLib)) != kOK)
    {
        std::cout << "Error: " << status << " - Failed to construct GoApiLib." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to a sensor system.
    kIpAddress systemIpAddress;
    kIpAddress_Parse(&systemIpAddress, SENSOR_IP);
    system = GoSystem(systemIpAddress, GO_PXL_SDK_DEFAULT_CONTROL_PORT);
    kChar ipAddress[16];
    kIpAddress_Format(system.Address(), ipAddress, sizeof(ipAddress));
    try
    {
        system.Connect();
        std::cout << "\nConnected to " << ipAddress << ":" << system.ControlPort() << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Connection failed. Check if sensor is powered on and connected "
                                 "also check if you are using the correct IP address and port number."
                  << std::endl;
        return ERROR_STATUS;
    }

    // Enable recording.
    payload = GoJson("{\"enabled\": true}");
    system.Client().Update(RECORDING_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
    std::cout << "\nRecording enabled." << std::endl;

    // Set frame rate.
    payload = GoJson("{\"parameters\" : {\"triggerSettings\" : {\"source\" : 0, \"frameRate\" : 50} }}");
    try
    {
        system.Client().Update(SENSOR_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
        std::cout << "\nFrame rate set to \"" << payload.At("parameters/triggerSettings/frameRate") << "\"." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Check if the API path is valid." << std::endl;
        return ERROR_STATUS;
    }

    // Run the sensor.
    if (system.RunningState() == GoSystem::State::Ready)
    {
        system.Start();
        std::cout << "\nSystem started." << std::endl;
    }

    // Wait to obtain data.
    kThread_Sleep(500'000);

    // Stop the sensor.
    system.Stop();
    std::cout << "\nSystem stopped." << std::endl;

    // Fetch data and Download as a file.
    payload = GoJson("{\"contents\" : [\"liveJob\", \"replay\"]}");

    // Get data.
    response = system.Client().Call(BACKUP_PATH, payload).GetResponse(RECEIVE_DATA_TIMEOUT).Payload();
    ByteArray data = response.At("data").GetBinary();
    std::cout << "\nRecording received." << std::endl;

    std::cout << "\nDownloading .gprec file..." << std::endl;
    try
    {
        // Download.
        char *buffer = (char *)data.data();
        std::ofstream recFile(RECORDING_FILE_PATH, std::ios::out | std::ios::binary);
        std::cout << "\nRecoding downloaded to " << std::filesystem::absolute(RECORDING_FILE_PATH) << std::endl;
        if (recFile.is_open())
        {
            recFile.write(buffer, data.size());
            recFile.close();
        }
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << "Failed to download the recording to " << std::filesystem::absolute(RECORDING_FILE_PATH) << std::endl;
        return ERROR_STATUS;
    }

    // Upload.
    payload = GoJson("{\"contents\" : [\"liveJob\", \"replay\"]}");
    std::ifstream recFile(RECORDING_FILE_PATH, std::ios::binary);
    std::vector<k8u> recByteArray;
    recByteArray = std::vector<k8u>(std::istreambuf_iterator<char>(recFile), {});
    payload.Set("data", recByteArray);

    std::cout << "\nUploading .gprec file..." << std::endl;
    try
    {
        system.Client().Call(RESTORE_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
        std::cout << "Uploaded recording file from " << std::filesystem::absolute(RECORDING_FILE_PATH) << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << "\nFailed to upload the recording. Make sure the file path is correct. Current file path: " << std::filesystem::absolute(RECORDING_FILE_PATH) << std::endl;
        return ERROR_STATUS;
    }

    // Enable replay mode.
    payload = GoJson("{\"enabled\" : true}");
    try
    {
        system.Client().Update(REPLAY_PLAYBACK_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
        std::cout << "\nReplay enabled." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << "\nFailed to enable recording. Check the API path." << std::endl;
        return ERROR_STATUS;
    }

    // Read the number of frames in the recording.
    response = system.Client().Read("/replay").GetResponse(REST_COMMAND_TIMEOUT).Payload();
    int maxFrameCount = response.Get<int>("frameCount");
    std::cout << "Number of frames in recording: " << maxFrameCount << std::endl;

    // Start playback.
    system.Client().Call(START_PATH).CheckResponse(REST_COMMAND_TIMEOUT);
    std::cout << "Playback started." << std::endl;

    // Stop playback.
    system.Client().Call(STOP_PATH).CheckResponse(REST_COMMAND_TIMEOUT);
    std::cout << "Playback stopped." << std::endl;

    // Jump to a specific frame.
    payload = GoJson();
    payload.Set("positionDomain", 0);
    payload.Set("position", maxFrameCount - 1);
    system.Client().Call(REPLAY_SEEK_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
    std::cout << "Framed changed to " << maxFrameCount - 1 << std::endl;

    system.Disconnect();
    return OK_STATUS;
}

int main(int argc, char **argv)
{
    int status;
    try
    {
        status = ReplayData();
    }
    catch (const GoRequestError &e)
    {
        string errorMessage = "";

        const GoJson &payload = e.GetResponse().Payload();
        if (!payload.Empty() && payload.HasKey("errors"))
        {
            GoJson errors = payload.At("errors");
            errorMessage.append(errors.At("/0/status").ToString());
            errorMessage.append(" ");
            errorMessage.append(errors.At("/0/description").ToString());
        }

        std::cerr << "GoRequestError: " << errorMessage << std::endl;
        std::cerr << "Error sending a REST command to " << e.GetResponse().Path() << std::endl;
        status = ERROR_STATUS;
    }
    catch (const GoChannelError &e)
    {
        std::cerr << "Error: " << e.what() << " - Check sensor status and make sure it is connected." << std::endl;
        status = ERROR_STATUS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        status = ERROR_STATUS;
    }

    std::cout << "\nPress ENTER to exit the program..." << std::endl;
    getchar();

    return status;
}
