/**
 * @file    GoGdpProfileBase.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
namespace GoPxLSdk.GoGdpMsg
{
    public abstract class GoGdpProfileBase : GoGdpMsg
    {
        /// <summary>Get number of points per profile array.</summary>
        /// <returns>profile width.</returns>
        public uint Width { get;  internal set; } = 0;

        /// <summary>Get number of values per profile intensity array.</summary>
        /// <returns>profile array width.</returns>
        public uint IntensityWidth { get; internal set; } = 0;

        /// <summary>Get profile resolution.</summary>
        /// <remarks>
        /// The actual resolution/offset is a 2-d value. 
        /// However, the Firesync 2-d structure kPoint64f only has X and Y members.
        /// The customer documentation refers to the resolution/scale and offset as Z scale and Z offset.
        /// To be consistent with the documentation, use a 3-d structure GoPoint3d64f that has a member called "Z", 
        /// instead of using the 2-d kPoint64f structure which does not have a "Z" member.
        /// The Y-member in the KPoint3d64f used for resolution and offset are set to 0.
        /// </remarks>
        /// <returns>resolution value.</returns>
        public GoPoint3d64f Resolution { get; internal set; } = new GoPoint3d64f(0, 0, 0);

        /// <summary>Get profile offset.</summary>
        /// <returns>offset value.</returns>
        public GoPoint3d64f Offset { get; internal set; } = new GoPoint3d64f(0, 0, 0);

        /// <summary>Get profile exposure.</summary>
        /// <returns>exposure value in nanoseconds.</returns>
        public float Exposure { get; internal set; }

        /// <summary>Constructs GoGdpProfileBase.</summary>
        /// <param name="type">The message type.</param>
        protected GoGdpProfileBase(MessageType type)
            : base(type)
        { }

        /// <summary>Deserialize information common to all profile messages.</summary>
        /// <param name="serializer">The serializer to read.</param>
        internal override void Deserialize(GoBinaryReader serializer)
        {
            base.Deserialize(serializer);

            serializer.BeginRead<ushort>(true);

            try
            {
                Width = serializer.ReadUInt32();
                IntensityWidth = serializer.ReadUInt32();

                var resolutionX = serializer.ReadDouble();
                var resolutionZ = serializer.ReadDouble();
                Resolution = new GoPoint3d64f(resolutionX, 0, resolutionZ);

                var offsetX = serializer.ReadDouble();
                var offsetZ = serializer.ReadDouble();
                Offset = new GoPoint3d64f(offsetX, 0, offsetZ);

                Exposure = serializer.ReadSingle();
            }
            finally
            {
                serializer.EndRead();
            }
        }
    }
}