/**
 * @file    GoGdpSurfaceBase.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
namespace GoPxLSdk.GoGdpMsg
{
    public class GoGdpSurfaceBase : GoGdpMsg
    {
        /// <summary>Get surface length (rows).</summary>
        /// <returns>surface length.</returns>
        public uint Length { get; internal set; }

        /// <summary>Get surface width (columns).</summary>
        /// <returns>surface width.</returns>
        public uint Width { get; internal set; }

        /// <summary>Get surface intensity length (rows).</summary>
        /// <returns>surface intensity length.</returns>
        public uint IntensityLength { get; internal set; }

        /// <summary>Get surface intensity width (columns).</summary>
        /// <returns>surface intensity width.</returns>
        public uint IntensityWidth { get; internal set; }

        /// <summary>Get surface scale point.</summary>
        /// <returns>surface resolution.</returns>
        public GoPoint3d64f Resolution { get; internal set; }

        /// <summary>Get surface offset point.</summary>
        /// <returns>surface offset.</returns>
        public GoPoint3d64f Offset { get; internal set; }

        /// <summary>Get surface id value which is changed whenever surface data is changed.</summary>
        /// <returns>surface id value.</returns>
        public uint SurfaceId { get; internal set; }

        /// <summary>Get surface exposure(ns).</summary>
        /// <returns>surface exposure value in nanoseconds.</returns>
        public float Exposure { get; internal set; }

        /// <summary>Constructs GoGdpSurfaceBase which forms the base class of the uniform surface and 
        /// surface point cloud messages.</summary>
        /// <param name="type">The message type.</param>
        protected GoGdpSurfaceBase(MessageType type)
            :base(type)
        { }

        /// <summary>Deserialize surface point cloud message.</summary>
        /// <param name="serializer">The serializer to read.</param>
        internal override void Deserialize(GoBinaryReader serializer)
        {
            base.Deserialize(serializer);

            serializer.BeginRead<ushort>(true);

            try
            {
                Length = serializer.ReadUInt32();
                Width = serializer.ReadUInt32();
                IntensityLength = serializer.ReadUInt32();
                IntensityWidth = serializer.ReadUInt32();

                Resolution = serializer.ReadGoPoint3d64f();
                Offset = serializer.ReadGoPoint3d64f();

                SurfaceId = serializer.ReadUInt32();
                Exposure = serializer.ReadSingle();

                // IMPORTANT!!!
                // Surface attributes serializer read section is still open here.
                // This is because the derived classes (ie. surface point cloud) has additional attributes
                // to read, so keep the section open for the derived class to finish reading
                // their specific attributes.

                // The read section must be closed by the derived classes.
            }
            catch
            {
                // Close read section on error.
                serializer.EndRead();
                throw;
            }
        }
    }
}