/**
 * @file    GoGdpTransform.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
namespace GoPxLSdk.GoGdpMsg
{
    /// <summary>
    /// This class represents a 3D transformation matrix
    /// [xx xy xz xt]
    /// [yx yy yz yt]
    /// [zx zy zz zt]
    /// [ 0  0  0  1]
    /// </summary>
    public class GoGdpTransform
    {
        /// <summary>Get the XX value of the 3D transformation matrix</summary>
        /// <returns>XX value of transform matrix</returns>
        public double XX { get; private set; } = 1.0;

        /// <summary>Get the XY value of the 3D transformation matrix</summary>
        /// <returns>XY value of transform matrix</returns>
        public double XY { get; private set; } = 0.0;

        /// <summary>Get the XZ value of the 3D transformation matrix</summary>
        /// <returns>XZ value of transform matrix</returns>
        public double XZ { get; private set; } = 0.0;

        /// <summary>Get the XT value of the 3D transformation matrix</summary>
        /// <returns>XT value of transform matrix</returns>
        public double XT { get; private set; } = 0.0;

        /// <summary>Get the YX value of the 3D transformation matrix</summary>
        /// <returns>YX value of transform matrix</returns>
        public double YX { get; private set; } = 0.0;

        /// <summary>Get the YY value of the 3D transformation matrix</summary>
        /// <returns>YY value of transform matrix</returns>
        public double YY { get; private set; } = 1.0;

        /// <summary>Get the YZ value of the 3D transformation matrix</summary>
        /// <returns>YZ value of transform matrix</returns>
        public double YZ { get; private set; } = 0.0;

        /// <summary>Get the YT value of the 3D transformation matrix</summary>
        /// <returns>YT value of transform matrix</returns>
        public double YT { get; private set; } = 0.0;

        /// <summary>Get the ZX value of the 3D transformation matrix</summary>
        /// <returns>ZX value of transform matrix</returns>
        public double ZX { get; private set; } = 0.0;

        /// <summary>Get the ZY value of the 3D transformation matrix</summary>
        /// <returns>ZY value of transform matrix</returns>
        public double ZY { get; private set; } = 0.0;

        /// <summary>Get the ZZ value of the 3D transformation matrix</summary>
        /// <returns>ZZ value of transform matrix</returns>
        public double ZZ { get; private set; } = 1.0;

        /// <summary>Get the ZT value of the 3D transformation matrix</summary>
        /// <returns>ZT value of transform matrix</returns>
        public double ZT { get; private set; } = 0.0;

        public GoGdpTransform() 
        { }

        public GoGdpTransform(double xX, double xY, double xZ, double xT, double yX, double yY, double yZ, double yT, double zX, double zY, double zZ, double zT)
        {
            XX = xX;
            XY = xY;
            XZ = xZ;
            XT = xT;

            YX = yX;
            YY = yY;
            YZ = yZ;
            YT = yT;

            ZX = zX;
            ZY = zY;
            ZZ = zZ;
            ZT = zT;
        }

        /// <summary>Deserialize the transform data.</summary>
        /// <param name="serializer">The serializer to read.</param>
        internal void Deserialize(GoBinaryReader serializer)
        {
            XX = (double)serializer.ReadSingle();
            XY = (double)serializer.ReadSingle();
            XZ = (double)serializer.ReadSingle();
            XT = (double)serializer.ReadSingle();

            YX = (double)serializer.ReadSingle();
            YY = (double)serializer.ReadSingle();
            YZ = (double)serializer.ReadSingle();
            YT = (double)serializer.ReadSingle();

            ZX = (double)serializer.ReadSingle();
            ZY = (double)serializer.ReadSingle();
            ZZ = (double)serializer.ReadSingle();
            ZT = (double)serializer.ReadSingle();
        }

        public override bool Equals(object obj)
        {
            return obj is GoGdpTransform transform &&
                   XX == transform.XX &&
                   XY == transform.XY &&
                   XZ == transform.XZ &&
                   XT == transform.XT &&
                   YX == transform.YX &&
                   YY == transform.YY &&
                   YZ == transform.YZ &&
                   YT == transform.YT &&
                   ZX == transform.ZX &&
                   ZY == transform.ZY &&
                   ZZ == transform.ZZ &&
                   ZT == transform.ZT;
        }

        public override int GetHashCode()
        {
            int hashCode = 1596574263;
            hashCode = hashCode * -1521134295 + XX.GetHashCode();
            hashCode = hashCode * -1521134295 + XY.GetHashCode();
            hashCode = hashCode * -1521134295 + XZ.GetHashCode();
            hashCode = hashCode * -1521134295 + XT.GetHashCode();
            hashCode = hashCode * -1521134295 + YX.GetHashCode();
            hashCode = hashCode * -1521134295 + YY.GetHashCode();
            hashCode = hashCode * -1521134295 + YZ.GetHashCode();
            hashCode = hashCode * -1521134295 + YT.GetHashCode();
            hashCode = hashCode * -1521134295 + ZX.GetHashCode();
            hashCode = hashCode * -1521134295 + ZY.GetHashCode();
            hashCode = hashCode * -1521134295 + ZZ.GetHashCode();
            hashCode = hashCode * -1521134295 + ZT.GetHashCode();
            return hashCode;
        }

        public static bool operator ==(GoGdpTransform a, GoGdpTransform b)
        {
            return a.Equals(b);
        }

        public static bool operator !=(GoGdpTransform a, GoGdpTransform b)
        {
            return !a.Equals(b);
        }
    }
}