/*
 * @file            ConfigureTool.cs
 * Objective        Connect to a Gocator sensor system, add a tool, change configurations of the tool, and enable measurement data output.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

using GoPxLSdk;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;

namespace ConfigureTool
{
    internal class ConfigureToolSample
    {
        private const string SENSOR_IP = "192.168.1.10";
        private const int PORT = 3600;
        private const string TOOLS_PATH = "/tools/";
        private const string DATA_SOURCE_PATH = "/dataSources/";
        private const string GOCATOR_CONTROL_PATH = "/controls/gocator";
        private const string GOCATOR_OUTPUT_PATH = "/controls/gocator/outputs/commands/add";
        private const int ERROR_STATUS = -1;
        private const int OK_STATUS = 0;
        private const int REST_COMMAND_TIMEOUT = 3_000; // 3 seconds.

        private static int ConfigureTool()
        {
            // Connect to a sensor system.
            IPAddress ipAddress;
            ipAddress = IPAddress.Parse(SENSOR_IP);
            using (var system = new GoSystem(ipAddress, PORT))
            {
                try
                {
                    system.Connect();
                    Console.WriteLine($"\nConnected to {ipAddress}:{PORT}");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(
                        ex.Message
                            + " - Connection failed. Check if sensor is powered on and connected "
                            + "also check if you are using the correct IP address and port number."
                    );
                    return ERROR_STATUS;
                }

                JObject response;
                JObject payload;

                // Configuration 1: Add a "Profile Bounding Box" tool.
                payload = new JObject { ["type"] = "ProfileBoundingBox", ["autoConnect"] = false, ["position"] = 0 };
                system.Client().Create(TOOLS_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine("\nNew tool added: " + payload["type"]);

                // Configuration 2: Update the tool id.
                payload = new JObject { ["extId"] = "ProfileBoundingBox-demo" };
                try
                {
                    system.Client().Update("tools/ProfileBoundingBox-0", payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("\nTool id updated: " + payload["extId"]);
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to update tool id. Check the existing tool id and try again.");
                    return ERROR_STATUS;
                }

                // Configuration 3: Connect tool to a data source.
                // Read all available outputs (data sources).
                response = system.Client().Read(DATA_SOURCE_PATH).GetResponse(REST_COMMAND_TIMEOUT).Payload;
                var dataSources = response.SelectToken("_embedded.item");

                string dataSourceId;
                var dataSourceIds = new List<string>();
                Console.WriteLine("\nData sources:");
                foreach (var it in dataSources)
                {
                    dataSourceId = it.SelectToken("_links.self.href").Value<string>();
                    Console.WriteLine(dataSourceId);
                    dataSourceIds.Add(dataSourceId);
                }
                // Read all available tool inputs.
                response = system.Client().Read("/tools/extId=ProfileBoundingBox-demo/inputs").GetResponse().Payload;
                var inputs = response.SelectToken("_embedded.item");

                string inputId;
                var inputIds = new List<string>();
                Console.WriteLine("\nTool inputs:");
                foreach (var it in inputs)
                {
                    inputId = it.SelectToken("_links.self.href").Value<string>();
                    Console.WriteLine(inputId);
                    inputIds.Add(inputId);
                }

                // Connect tool input to a data source.
                string profileInputPath = "/tools/extId=ProfileBoundingBox-demo/inputs/ProfileInput";
                payload = new JObject { ["dataSource"] = "scan:g2xxx:scanner-0:topUniformProfile" };
                try
                {
                    system.Client().Update(profileInputPath, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("\nTool input connected.");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to update tool input. Check the tool id and API path.");
                    return ERROR_STATUS;
                }

                // Configuration 4: Enable a measurement output.
                payload = new JObject { ["enabled"] = true };
                string output = "X";
                string MEASUREMENT_OUTPUT_PATH = "/tools/extId=ProfileBoundingBox-demo/outputs/" + output;
                try
                {
                    system.Client().Update(MEASUREMENT_OUTPUT_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("\nMeasurement output enabled.");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to enable the output. Check the tool id and API path.");
                    return ERROR_STATUS;
                }

                // Configuration 5: Send measurement value over GDP.
                // Read measurement output.
                string METRICS_PATH = "/tools/extId=ProfileBoundingBox-demo/metrics";

                try
                {
                    response = system.Client().Read(METRICS_PATH).GetResponse().Payload;
                    Console.WriteLine("\nMeasurement value: " + response["/outputsByExtId/" + output + "/value"]);
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to read value. Check the API path and ensure scan data is available.");
                    return ERROR_STATUS;
                }

                // Enable Gocator Protocol.
                payload = new JObject { ["enabled"] = true };
                system.Client().Update(GOCATOR_CONTROL_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine("\nGocator Protocol enabled.");

                // Send the value.
                payload = new JObject { ["source"] = "tools:ProfileBoundingBox-0:outputs:X", ["outputId"] = 1, ["autoShift"] = true };
                try
                {
                    system.Client().Call(GOCATOR_OUTPUT_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("\nMeasurement value sent over GDP.");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to send the data. Check the tool name and API path.");
                    return ERROR_STATUS;
                }

                system.Disconnect();
                return OK_STATUS;

            }
        }

        static int Main()
        {
            int status;

            try
            {
                status = ConfigureTool();
            }
            catch (GoRequestException ex)
            {
                string errorMessage = "";

                JObject payload = ex.Response.Payload;
                if (!(payload == null) && !payload.HasValues && payload.ContainsKey("errors"))
                {
                    JToken error = ex.Response.Payload["errors"][0];
                    errorMessage = error != null ? $"{error["status"]} {error["description"]}" : "";
                }

                Console.WriteLine("GoRequestError: " + errorMessage);
                Console.WriteLine("Error sending a REST command to " + ex.Response.Path);
                status = ERROR_STATUS;
            }
            catch (GoChannelException ex)
            {
                Console.WriteLine($"Error: {ex.Message} - Check sensor status and make sure it is connected.");
                status = ERROR_STATUS;
            }
            catch (Exception ex)
            {
                Console.WriteLine($"Error: {ex.Message}");
                status = ERROR_STATUS;
            }

            Console.WriteLine("\nPress any key to exit the program...");
            Console.ReadKey(true);
            return status;
        }
    }
}