/*
 * @file            Discover.cs
 * Objective        Discover sensor and GoPxL instances on the network.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

using GoPxLSdk;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.NetworkInformation;

namespace Discover
{
    internal class DiscoverSample
    {
        private const string ENV_CONFIG_PATH = "/environ/ipConfig";
        private const string VISIBLE_SENSORS_PATH = "/scan/visibleSensors/";
        private const int ERROR_STATUS = -1;
        private const int OK_STATUS = 0;
        private const int DISCOVER_TIMEOUT = 3_000; // 3 seconds.
        
        private static int Discover()
        {
            // Discover sensors.
            var discovery = new GoDiscoveryClient();
            discovery.BlockingDiscover(DISCOVER_TIMEOUT);
            List<GoInstance> instances = discovery.InstanceList();
            Console.WriteLine($"Number of sensors on the network: {instances.Count}");

            if (instances.Count == 0)
            {
                Console.WriteLine("No sensors found, make sure sensors or GoPxl on PC/GoMax are available and connected.");
                return ERROR_STATUS;
            }

            // Iterate over the sensors to get their information.
            int instanceCount = 0;
            foreach (GoInstance instance in instances)
            {
                Console.WriteLine();
                Console.WriteLine(new string('*', 52) + " GoPxL instance " + (instanceCount + 1)); // Prints 52 asterisks to help the readability.
                Console.WriteLine("IP Address: " + instance.IpAddress);
                Console.WriteLine("Gateway: " + instance.Gateway);
                Console.WriteLine("Mask: " + instance.Mask);
                Console.WriteLine("DHCP: " + (instance.IsDhcp ? "True" : "False"));
                Console.WriteLine("Application ID: " + instance.AppId);
                Console.WriteLine("Application Name: " + instance.AppName);
                Console.WriteLine("Application Version: " + instance.AppVersion);
                Console.WriteLine("Control Port: " + instance.ControlPort);
                Console.WriteLine("GDP Port: " + instance.GdpPort);
                Console.WriteLine("Web Port: " + instance.WebPort);
                Console.WriteLine("IP Conflict: " + (instance.IsAddressConflict ? "Yes" : "No"));
                ++instanceCount;

                // Connect to a sensor/instance.
                using (var system = new GoSystem(instance.IpAddress, (int)instance.ControlPort))
                {
                    try
                    {
                        system.Connect();
                        Console.WriteLine($"\nConnected to {instance.IpAddress}:{instance.ControlPort}");
                    }
                    catch (GoRequestException ex)
                    {
                        Console.WriteLine(ex.Message + " - Connection failed.");
                        return ERROR_STATUS;
                    }
                    try
                    {
                        // If this is a multi-sensor system, iterate to get each sensor information.
                        JToken[] sensors = system.Client().Read(VISIBLE_SENSORS_PATH).GetResponse().Payload["sensors"].ToArray();
                        int sensorCount = 0;
                        foreach (JToken sensor in sensors)
                        {
                            var sensorSerialNumber = sensor["serialNumber"].Value<string>();
                            var sensorPath = system.SensorPath(sensorSerialNumber);
                            if (!string.IsNullOrEmpty(sensorPath))
                            {
                                JToken ipConfig = system.Client().Read(ENV_CONFIG_PATH).GetResponse().Payload["interfaces"];

                                Console.WriteLine();
                                Console.WriteLine(new string('*', 22) + " Sensor " + (sensorCount + 1));
                                Console.WriteLine("Local sensor: " + (sensor["isLocal"].Value<bool>() ? "Yes" : "No"));
                                Console.WriteLine("IP Address: " + sensor.SelectToken("parameters.ipAddress").Value<string>());
                                Console.WriteLine("Gateway Address: " + ipConfig.SelectToken("[0].gatewayAddress").Value<string>());
                                Console.WriteLine("Subnet Mask: " + ipConfig.SelectToken("[0].subnetMask").Value<string>());
                                Console.WriteLine("DHCP: " + ipConfig.SelectToken("[0].dhcp"));
                                Console.WriteLine("Serial Number: " + sensorSerialNumber);
                                Console.WriteLine("Model: " + sensor["model"].Value<string>());
                                Console.WriteLine("Engine ID: " + sensor["engineId"].Value<string>());
                                Console.WriteLine("Sensor Path: " + sensorPath);
                            }
                            else
                            {
                                continue;
                            }
                            ++sensorCount;
                        }
                    }
                    catch (GoRequestException ex)
                    {
                        Console.WriteLine(ex.Message + " - Failed to connect. Sensor might be controlled by another instance.");
                    }
                }
            }
            return OK_STATUS;
        }

        static int Main()
        {
            int status = Discover();

            Console.WriteLine("\nPress any key to exit the program...");
            Console.ReadKey(true);

            return status;
        }
    }
}