/*
 * @file            ReceiveMeasurement.cs
 * Objective        Connect to the GDP server and receive measurement data.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 *
 * Note: Tool measurement output must be enabled and available. Refer
 * to ConfigureTool.cs for enabling the measurement outputs.
 */

using GoPxLSdk;
using GoPxLSdk.GoGdpMsg;
using Newtonsoft.Json.Linq;
using System;
using System.Net;

namespace ReceiveMeasurement
{
    internal class ReceiveMeasurementSample
    {
        private const string SENSOR_IP = "192.168.1.10";
        private const int PORT = 3600;
        private const string GOCATOR_CONTROL_PATH = "/controls/gocator";
        private const int ERROR_STATUS = -1;
        private const int OK_STATUS = 0;
        private const int REST_COMMAND_TIMEOUT = 3_000; // 3 seconds.
        private const int RECEIVE_DATA_TIMEOUT = 3_000;

        private const int GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT = 3601;
        private static int ReceiveMeasurement()
        {
            // Connect to a sensor system.
            IPAddress ipAddress;
            ipAddress = IPAddress.Parse(SENSOR_IP);
            using (var system = new GoSystem(ipAddress, PORT))
            {
                try
                {
                    system.Connect();
                    Console.WriteLine($"\nConnected to {ipAddress}:{PORT}");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(
                        ex.Message
                            + " - Connection failed. Check if sensor is powered on and connected "
                            + "also check if you are using the correct IP address and port number."
                    );
                    return ERROR_STATUS;
                }

                JObject payload;

                // Run the sensor.
                if (system.RunningState() == GoSystem.State.Ready)
                {
                    system.Start();
                    Console.WriteLine("\nSystem started.");
                }

                // Enable Gocator Protocol.
                payload = new JObject { ["enabled"] = true };
                try
                {
                    system.Client().Update(GOCATOR_CONTROL_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("\nGocator Protocol enabled.");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Check the timeout value or the API path.");
                    return ERROR_STATUS;
                }

                // Connect to the Gocator Data Protocol (GDP) server.
                using (var gdpClient = new GoGdpClient())
                {
                    try
                    {
                        gdpClient.Connect(system.Address, GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT);
                        Console.WriteLine("\nGocator Protocol Connected.");
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(ex.Message + " - Failed to connect to the GDP server.");
                        return ERROR_STATUS;
                    }

                    // Establish a synchronous data transfer.
                    try
                    {
                        gdpClient.ReceiveDataSync(RECEIVE_DATA_TIMEOUT);
                        Console.WriteLine("\nSynchronous data transfer established.");
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(ex.Message + " - Failed to receive data via synchronous connection.");
                        return ERROR_STATUS;
                    }

                    // Receive data.
                    GoDataSet receivedDataSet = gdpClient.DataSet;
                    if (receivedDataSet.Count < 1)
                    {
                        Console.WriteLine("\nNo data is received.");
                        return ERROR_STATUS;
                    }
                    Console.WriteLine("\nTotal number of messages: " + receivedDataSet.Count);

                    // Loop through messages in the dataSet.
                    for (int i = 0; i < receivedDataSet.Count; i++)
                    {
                        Console.WriteLine($"\nMessage {i + 1} out of {receivedDataSet.Count}");
                        Console.WriteLine($"Data source: {receivedDataSet.GdpMsgAt(i).DataSourceId}");

                        Console.WriteLine($"type: {receivedDataSet.GdpMsgAt(i).Type}");

                        switch (receivedDataSet.GdpMsgAt(i).Type)
                        {
                            // Get measurement output.
                            case MessageType.Measurement:
                                {
                                    GoGdpMeasurement measurementMsg = (GoGdpMeasurement)receivedDataSet.GdpMsgAt(i);
                                    Console.WriteLine("Array count: " + measurementMsg.ArrayedCount);
                                    Console.WriteLine("Array index: " + measurementMsg.ArrayedIndex);
                                    Console.WriteLine("Value: " + measurementMsg.Value);
                                    Console.WriteLine("Decision: " + (int)measurementMsg.Decision);
                                    Console.WriteLine("Output ID: " + measurementMsg.GdpId);
                                    break;
                                }
                        }
                    }

                    Console.WriteLine("\nPress ENTER to exit the program...");
                    Console.ReadKey(true);
                    gdpClient.Close();
                }

                system.Stop();
                Console.WriteLine("\nSystem stopped.");
                system.Disconnect();
            }

            return OK_STATUS;
        }

        static int Main()
        {
            int status = ReceiveMeasurement();
            return status;
        }
    }
}