/*
 * @file            ReceiveSurface.cs
 * Objective        Connect to the GoPxL Data Protocol and receive surface data.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 *
 * GoGdpClient::ReceiveDataSync() or GoGdpClient::ReceiveDataAsync() can be used to receive data from the GoPxL Data Protocol (GDP) serve:
    1. Pass a timeout value (in microseconds) within which data will be received from the server.
    2. The GDP server sends out individual data messages. The last message is marked with an isLastMsg flag.
    3. ReceiveDataSync() receives individual data messages and reassembles to a data set (GoDataSet).
    4. GoGdpClient::DataSet() can be used to access the data set.
    5. GoGdpMsg::Type() indicates the type of individual messages in the dataset as described in GoPxLSdk::MessageType.
 */

using GoPxLSdk;
using GoPxLSdk.GoGdpMsg;
using System;
using System.Diagnostics;
using System.Net;
using Newtonsoft.Json.Linq;
using System.Threading.Tasks;
using System.Data;
using System.Collections.Generic;

namespace ReceiveSurface
{
    internal class ReceiveSurfaceSample
    {
        private const string SENSOR_IP = "192.168.1.10";
        private const int PORT = 3600;
        private const string ENGINE_ID = "g2xxx";
        private const string OUTPUTS_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/outputs";
        private const string OUTPUT_UNIFORM_SURFACE_PATH = OUTPUTS_PATH + "/topUniformSurface";
        private const string OUTPUT_POINT_CLOUD_SURFACE_PATH = OUTPUTS_PATH + "/sensor-0PointcloudSurface";
        private const string GOCATOR_CONTROL_PATH = "/controls/gocator";
        private const string GOCATOR_ADD_OUTPUT_PATH = "/controls/gocator/outputs/commands/add";
        private const int ERROR_STATUS = -1;
        private const int OK_STATUS = 0;
        private const int REST_COMMAND_TIMEOUT = 3_000; // 3 seconds.
        private const int RECEIVE_DATA_TIMEOUT = 3_000;

        private const int GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT = 3601;

        private struct SurfacePoint
        {
            public double X { get; set; }   // X coordinate - position along laser line (mm).
            public double Y { get; set; }   // Y coordinate - position along the direction of travel (mm).
            public double Z { get; set; }   // Z coordinate - height at a given X position(mm).
            public byte Intensity { get; set; }
        }

        private struct UniformSurface
        {
            /* The X and Y coordinates of a given point in the height map can be calculated as following:
             * for Z value of HeightMap[i][j]
             * double X = XOffset + XResolution * i;
             * double X = YOffset + YResolution * j;
             */

            public UniformSurface(uint length, uint width, double XOffset, double YOffset, double XResolution, double YResolution)
            {
                heightMap = new double[length][];
                for (uint i = 0; i < length; i++)
                {
                    heightMap[i] = new double[width];
                }
                this.XOffset = XOffset;
                this.YOffset = YOffset;
                this.XResolution = XResolution;
                this.YResolution = YResolution;
            }

            public double[][] heightMap; // Z coordinates - height at a given X and Y position (mm).
            public readonly double XOffset; // X Offset - offset position along laser line (mm).
            public readonly double YOffset; // Y Offset - offset position along the direction of travel (mm).
            public readonly double XResolution; // X Resolution - resolution along laser line (mm).
            public readonly double YResolution; // Y resolution - resolution along the direction of travel (mm).
        }

        private static int ReceiveSurfaceNet()
        {
            // Connect to a sensor system.
            IPAddress ipAddress;
            ipAddress = IPAddress.Parse(SENSOR_IP);
            using (var system = new GoSystem(ipAddress, PORT))
            {
                try
                {
                    system.Connect();
                    Console.WriteLine($"\nConnected to {ipAddress}:{PORT}");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(
                        ex.Message
                            + " - Connection failed. Check if sensor is powered on and connected "
                            + "also check if you are using the correct IP address and port number."
                    );
                    return ERROR_STATUS;
                }

                JObject response;
                JObject payload;
                GoDataSet receivedDataSet;

                // Run the sensor.
                if (system.RunningState() == GoSystem.State.Ready)
                {
                    system.Start();
                    Console.WriteLine("\nSystem started.");
                }

                // Enable the Gocator Protocol.
                payload = new JObject { ["enabled"] = true };
                try
                {
                    system.Client().Update(GOCATOR_CONTROL_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("\nGocator Protocol enabled.");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Check the timeout value or the API path.");
                    return ERROR_STATUS;
                }

                // Add uniform surface to Gocator Protocol output.
                payload = new JObject { ["enabled"] = true };
                try
                {
                    response = system.Client().Read(OUTPUT_POINT_CLOUD_SURFACE_PATH).GetResponse().Payload;
                    //response = system.Client().Read(OUTPUT_UNIFORM_SURFACE_PATH).GetResponse().Payload;
                    string surfaceDataSourceId = response["dataSourceId"].ToString();
                    payload = new JObject { ["source"] = surfaceDataSourceId, ["outputId"] = 0, ["autoShift"] = true };

                    system.Client().Call(GOCATOR_ADD_OUTPUT_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("\nAdded surface to the Gocator Protocol output.");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to add surface data to Gocator Protocol output. Check if uniform/point cloud toggle is correctly set.");
                    return ERROR_STATUS;
                }

                // Connect to the Gocator Data Protocol (GDP) server.
                using (var gdpClient = new GoGdpClient())
                {
                    try
                    {
                        gdpClient.Connect(system.Address, GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT);
                        Console.WriteLine("\nGocator Protocol Connected.");
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(ex.Message + " - Failed to connect to the GDP server.");
                        return ERROR_STATUS;
                    }

                    // Establish a synchronous data transfer.
                    bool surfaceReceived = false;
                    var timeout = TimeSpan.FromSeconds(10);
                    Stopwatch watch = new Stopwatch();
                    Console.WriteLine("\nWaiting until a dataset with Surface data is received within 10 seconds.");

                    watch.Start();
                    // Repeat receiving data within the specified timeout until a Surface data is received.
                    while (!surfaceReceived && watch.Elapsed < timeout)
                    {
                        // Establish a synchronous data transfer with a 3-second timeout.
                        try
                        {
                            gdpClient.ReceiveDataSync(RECEIVE_DATA_TIMEOUT);
                            Console.WriteLine("\nSynchronous data transfer established.");
                        }
                        catch (Exception ex)
                        {
                            Console.WriteLine("Error: " + ex.Message + " - Failed to receive data via synchronous connection.");
                            return ERROR_STATUS;
                        }

                        Console.Write("."); // Output a dot indicating loop is still running.

                        // Receive data.
                        receivedDataSet = gdpClient.DataSet;
                        if (receivedDataSet.Count < 1)
                        {
                            Console.WriteLine("\nNo data is received.");
                            return ERROR_STATUS;
                        }
                        Console.WriteLine("\nTotal number of messages: " + receivedDataSet.Count);

                        // Loop through messages in the dataSet.
                        for (int i = 0; i < receivedDataSet.Count; i++)
                        {
                            if (receivedDataSet.GdpMsgAt(i).Type == MessageType.UniformSurface ||
                                receivedDataSet.GdpMsgAt(i).Type == MessageType.SurfacePointCloud)
                            {
                                Console.WriteLine("\nA dataset with a Surface message has been received.");
                                surfaceReceived = true;
                                break;
                            }
                        }
                    }

                    // Get the data set again so that it is available outside the while loop scope.
                    receivedDataSet = gdpClient.DataSet;

                    Console.WriteLine("\nTotal number of messages in this data set: " + receivedDataSet.Count);
                    // Loop through messages in the dataSet.
                    for (int i = 0; i < receivedDataSet.Count; i++)
                    {
                        Console.WriteLine($"\nMessage {i + 1} out of {receivedDataSet.Count}");
                        Console.WriteLine($"Message Type {(int)receivedDataSet.GdpMsgAt(i).Type}: {receivedDataSet.GdpMsgAt(i).DataSourceId}");

                        switch (receivedDataSet.GdpMsgAt(i).Type)
                        {
                            case MessageType.Stamp:
                                {
                                    var stampMsg = (GoGdpStamp)receivedDataSet.GdpMsgAt(i);
                                    Console.WriteLine("\nStamp");
                                    Console.WriteLine("Frame index: " + stampMsg.FrameIndex);
                                    Console.WriteLine("Timestamp: " + stampMsg.Timestamp);
                                    Console.WriteLine("Encoder: " + stampMsg.Encoder);
                                    Console.WriteLine("Encoder at Z: " + stampMsg.EncoderAtZ);
                                    Console.WriteLine("Status: " + stampMsg.Status);
                                    break;
                                }
                            case MessageType.UniformSurface:
                                {
                                    GoGdpSurfaceUniform uniformSurfaceMsg = (GoGdpSurfaceUniform)receivedDataSet.GdpMsgAt(i);
                                    Console.WriteLine("\nUniform Surface");
                                    Console.WriteLine("Date Source: " + uniformSurfaceMsg.DataSourceId);
                                    Console.WriteLine("Array count: " + uniformSurfaceMsg.ArrayedCount);
                                    Console.WriteLine("Array index: " + uniformSurfaceMsg.ArrayedIndex);

                                    // Transform to real world coordinates.
                                    var length = uniformSurfaceMsg.Length;
                                    var width = uniformSurfaceMsg.Width;
                                    var intensityArray = uniformSurfaceMsg.Intensities;
                                    var intensityLength = uniformSurfaceMsg.IntensityLength;
                                    var intensityWidth = uniformSurfaceMsg.IntensityWidth;

                                    UniformSurface uniformSurface = new UniformSurface(length, width,
                                        uniformSurfaceMsg.Offset.X,
                                        uniformSurfaceMsg.Offset.Y,
                                        uniformSurfaceMsg.Resolution.X,
                                        uniformSurfaceMsg.Resolution.Y);

                                    /*  The surface data is sent in a 2D array of z values.
                                     *  The x and y real-world coordinates can be calculated from the indexes of the 2D array,
                                     *  using the x and y offsets and resolutions.
                                     *  The z values must be transformed into real world coordinates before being used.
                                     */
                                    for (int j = 0; j < length; j++)
                                    {
                                        for (int k = 0; k < width; k++)
                                        {
                                            short data = uniformSurfaceMsg.Ranges[j, k];

                                            // -32768 represents invalid data
                                            if (data != short.MinValue)
                                            {
                                                // Calculates the Z position in mm from the given data.
                                                // Z = Z offset + Z resolution * z
                                                uniformSurface.heightMap[j][k] = uniformSurfaceMsg.Offset.Z + uniformSurfaceMsg.Resolution.Z * data;
                                            }
                                            else
                                            {
                                                uniformSurface.heightMap[j][k] = short.MinValue;
                                            }
                                        }
                                    }
                                    Console.WriteLine("Uniform surface data length: " + length);
                                    Console.WriteLine("Uniform surface data width: " + width);

                                    Console.WriteLine((intensityArray != null ? "Intensity is enabled." : "Intensity is disabled."));
                                    if (intensityArray != null)
                                    {
                                        Console.WriteLine("Intensity length: " + intensityLength);
                                        Console.WriteLine("Intensity width: " + intensityWidth);
                                    }
                                    break;
                                }

                            case MessageType.SurfacePointCloud:
                                {

                                    GoGdpSurfacePointCloud pointCloudSurfaceMsg = (GoGdpSurfacePointCloud)receivedDataSet.GdpMsgAt(i);
                                    Console.WriteLine("\nPoint Cloud Surface");
                                    Console.WriteLine("Date Source: " + pointCloudSurfaceMsg.DataSourceId);
                                    Console.WriteLine("Array count: " + pointCloudSurfaceMsg.ArrayedCount);
                                    Console.WriteLine("Array index: " + pointCloudSurfaceMsg.ArrayedIndex);

                                    // Transform to real world coordinates.
                                    var length = pointCloudSurfaceMsg.Length;
                                    var width = pointCloudSurfaceMsg.Width;
                                    var intensityArray = pointCloudSurfaceMsg.Intensities;
                                    var intensityLength = pointCloudSurfaceMsg.IntensityLength;
                                    var intensityWidth = pointCloudSurfaceMsg.IntensityWidth;

                                    var surfaceBuffer = new SurfacePoint[length, width];

                                    for (int j = 0; j < length; j++)
                                    {
                                        for (int k = 0; k < width; k++)
                                        {
                                            GoPoint3d16s point = pointCloudSurfaceMsg.Ranges[j, k];

                                            surfaceBuffer[j, k].X = pointCloudSurfaceMsg.Offset.X + pointCloudSurfaceMsg.Resolution.X * point.X;
                                            surfaceBuffer[j, k].Y = pointCloudSurfaceMsg.Offset.Y + pointCloudSurfaceMsg.Resolution.Y * point.Y;

                                            // -32768 represents invalid data
                                            if (point.Z != short.MinValue)
                                            {
                                                surfaceBuffer[j, k].Z = pointCloudSurfaceMsg.Offset.Z + pointCloudSurfaceMsg.Resolution.Z * point.Z;
                                            }
                                            else
                                            {
                                                surfaceBuffer[j, k].Z = short.MinValue;
                                            }
                                        }
                                    }

                                    Console.WriteLine("Surface data length: " + length);
                                    Console.WriteLine("Surface data width: " + width);

                                    Console.WriteLine((intensityArray != null ? "Intensity is enabled." : "Intensity is disabled."));
                                    if (intensityArray != null)
                                    {
                                        Console.WriteLine("Intensity length: " + intensityLength);
                                        Console.WriteLine("Intensity width: " + intensityWidth);
                                    }
                                    break;
                                }
                        }
                    }
                    gdpClient.Close();
                }

                system.Stop();
                Console.WriteLine("\nSystem stopped.");
                system.Disconnect();
            }

            return OK_STATUS;
        }

        static int Main()
        {
            int status = ReceiveSurfaceNet();

            Console.WriteLine("\nPress any key to exit the program...");
            Console.ReadKey(true);

            return status;
        }
    }
}