/*
 * @file            ReplayData.cs
 * Objective        Connect to a Gocator sensor system, record profile scans and download/upload a recording file (.gprec).
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

using GoPxLSdk;
using Newtonsoft.Json.Linq;
using System;
using System.Collections;
using System.IO;
using System.Net;
using System.Xml.Linq;

namespace ReplayData
{
    internal class ReplayDataSample
    {
        private const string SENSOR_IP = "192.168.1.10";
        private const int PORT = 3600;
        private const string ENGINE_ID = "g2xxx";
        private const string SENSOR_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0";
        private const string START_PATH = "/system/commands/start";
        private const string STOP_PATH = "/system/commands/stop";
        private const string RECORDING_PATH = "/replay/recording";
        private const string BACKUP_PATH = "/system/commands/archive";
        private const string RESTORE_PATH = "/system/commands/restore";
        private const string REPLAY_SEEK_PATH = "/replay/commands/seek";
        private const string REPLAY_PLAYBACK_PATH = "/replay/playback";
        private const string RECORDING_FILE_PATH = "./sample_recording.gprec";
        private const int ERROR_STATUS = -1;
        private const int OK_STATUS = 0;
        private const int REST_COMMAND_TIMEOUT = 3_000; // 3 seconds.
        private const int RECEIVE_DATA_TIMEOUT = 20_000; // Receiving data might require a higher timeout value.

        private static int ReplayData()
        {
            // Connect to a sensor system.
            IPAddress ipAddress;
            ipAddress = IPAddress.Parse(SENSOR_IP);
            using (var system = new GoSystem(ipAddress, PORT))
            {
                try
                {
                    system.Connect();
                    Console.WriteLine($"\nConnected to {ipAddress}:{PORT}");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(
                        ex.Message
                            + " - Connection failed. Check if sensor is powered on and connected "
                            + "also check if you are using the correct IP address and port number."
                    );
                    return ERROR_STATUS;
                }

                JObject response;
                JObject payload;

                // Enable recording. 
                payload = new JObject { ["enabled"] = true };
                system.Client().Update(RECORDING_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine("\nRecording enabled.");

                // Set frame rate. 
                payload = new JObject { ["parameters"] = new JObject { ["triggerSettings"] = new JObject { ["source"] = 0, ["frameRate"] = 50 } } };
                try
                {
                    system.Client().Update(SENSOR_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("\nFrame rate set to \"" + payload.SelectToken("parameters.triggerSettings.frameRate") + "\".");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Check if the API path is valid.");
                    return ERROR_STATUS;
                }

                // Run the sensor.
                if (system.RunningState() == GoSystem.State.Ready)
                {
                    system.Start();
                    Console.WriteLine("\nSystem started.");
                }

                // Wait to obtain data. 
                System.Threading.Thread.Sleep(500);

                // Stop the sensor.
                system.Stop();
                Console.WriteLine("\nSystem stopped.");

                // Fetch data and Download as a file. 
                payload = new JObject { ["contents"] = new JArray() { "liveJob", "replay" } };

                // Get data.
                response = system.Client().Call(BACKUP_PATH, payload).GetResponse(RECEIVE_DATA_TIMEOUT).Payload;
                byte[] data = response["data"].ToObject<byte[]>();
                Console.WriteLine("\nRecording received.");

                Console.WriteLine("\nDownloading .gprec file...");
                try
                {
                    File.WriteAllBytes(RECORDING_FILE_PATH, data);
                    Console.WriteLine("\nRecoding downloaded to " + Path.GetFullPath(RECORDING_FILE_PATH));
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to download the recording to " + Path.GetFullPath(RECORDING_FILE_PATH));
                    return ERROR_STATUS;
                }

                // Upload.
                var recByteArray = File.ReadAllBytes(RECORDING_FILE_PATH);
                payload = new JObject { ["contents"] = new JArray() { "liveJob", "replay" }, ["data"] = data };
                Console.WriteLine("\nUploading .gprec file...");
                try
                {
                    system.Client().Call(RESTORE_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("Uploaded recording file from " + Path.GetFullPath(RECORDING_FILE_PATH));
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message
                        + " - Failed to upload the recording. Make sure the file path is correct. Current file path: "
                        + Path.GetFullPath(RECORDING_FILE_PATH));
                    return ERROR_STATUS;
                }

                // Enable replay mode. 
                payload = new JObject { ["enabled"] = true };
                try
                {
                    system.Client().Update(REPLAY_PLAYBACK_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("\nReplay enabled.");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to enable recording. Check the API path.");
                    return ERROR_STATUS;
                }

                // Read the number of frames in the recording.
                response = system.Client().Read("/replay").GetResponse(REST_COMMAND_TIMEOUT).Payload;
                int maxFrameCount = response["frameCount"].Value<int>();
                Console.WriteLine("Number of frames in recording: " + maxFrameCount);

                // Start playback.
                system.Client().Call(START_PATH).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine("Playback started.");

                // Stop playback.
                system.Client().Call(STOP_PATH).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine("Playback stopped.");

                // Jump to a specific frame.
                payload = new JObject { ["positionDomain"] = 0, ["position"] = maxFrameCount - 1 };
                system.Client().Call(REPLAY_SEEK_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine("Framed changed to " + (maxFrameCount - 1));

                system.Disconnect();
                return OK_STATUS;
            }
        }

        static int Main()
        {
            int status = ReplayData();

            Console.WriteLine("\nPress any key to exit the program...");
            Console.ReadKey(true);

            return status;
        }
    }
}